import json
import re
import typing as t

import gql_alchemy.query_model as qm
from .errors import GqlSchemaError
from .utils import PrimitiveType


class _SchemaAssertionError(RuntimeError):
    def __init__(self, message: str, name: str) -> None:
        super().__init__(message, name)

    def __str__(self) -> str:
        return t.cast(str, self.args[0])


class TypeResolvingError(RuntimeError):
    def __init__(self, name: str) -> None:
        super().__init__("Can not resolve `{}` type".format(name), name)

    def __str__(self) -> str:
        return t.cast(str, self.args[0])


class UndefinedVariableError(RuntimeError):
    def __init__(self, name: str) -> None:
        super().__init__("Undefined variable `{}`".format(name), name)

    def __str__(self) -> str:
        return t.cast(str, self.args[0])


class NonCompatibleVariableType(RuntimeError):
    def __init__(self, var_name: str, expected: 'GqlType', got: 'GqlType') -> None:
        super().__init__(
            "Type of `{}` variable is not compatible, `{}` expected, but got `{}`".format(var_name, expected, got),
            var_name, expected, got)

    def __str__(self) -> str:
        return t.cast(str, self.args[0])


class GqlType:
    def __init__(self, name: str) -> None:
        self.__name = name

    def is_assignable(self, value: PrimitiveType, type_registry: 'TypeRegistry') -> bool:
        raise NotImplementedError()

    def __str__(self) -> str:
        return self.__name


class _PossibleInputType(GqlType):
    def validate_input(self, value: t.Union[qm.Value, qm.ConstValue],
                       vars_values: t.Optional[t.Mapping[str, PrimitiveType]],
                       vars_defs: t.Mapping[str, GqlType],
                       type_registry: 'TypeRegistry') -> bool:
        raise NotImplementedError()

    def is_assignable(self, value: PrimitiveType, type_registry: 'TypeRegistry') -> bool:
        raise NotImplementedError()


class _TypeWrapper(_PossibleInputType):
    def __init__(self, of_type: t.Union['InlineType', str]) -> None:
        if isinstance(of_type, GqlType):
            of_type_name = str(of_type)
        else:
            of_type_name = of_type

        super().__init__(self._create_name(of_type_name))

        self.__of_type = of_type
        self.__of_type_resolved: t.Optional[GqlType] = None

    def of_type(self, type_registry: 'TypeRegistry') -> GqlType:
        if self.__of_type_resolved is None:
            self.__of_type_resolved = type_registry.resolve_type(self.__of_type)
        return self.__of_type_resolved

    def is_assignable(self, value: PrimitiveType, type_registry: 'TypeRegistry') -> bool:
        raise NotImplementedError()

    def validate_input(self, value: t.Union[qm.Value, qm.ConstValue],
                       vars_values: t.Optional[t.Mapping[str, PrimitiveType]],
                       vars_defs: t.Mapping[str, GqlType],
                       type_registry: 'TypeRegistry') -> bool:
        raise NotImplementedError()

    def _create_name(self, of_type: str) -> str:
        raise NotImplementedError()


class List(_TypeWrapper):
    def _create_name(self, of_type: str) -> str:
        return "[" + of_type + ']'

    def is_assignable(self, value: PrimitiveType, type_registry: 'TypeRegistry') -> bool:
        if value is None:
            return True

        if not isinstance(value, list):
            return False

        wrapped_type = self.of_type(type_registry)

        for el in value:
            if not wrapped_type.is_assignable(el, type_registry):
                return False

        return True

    def validate_input(self, value: t.Union[qm.Value, qm.ConstValue],
                       vars_values: t.Optional[t.Mapping[str, PrimitiveType]],
                       vars_defs: t.Mapping[str, GqlType],
                       type_registry: 'TypeRegistry') -> bool:
        if isinstance(value, qm.Variable):
            _validate_variable(value, self, vars_defs)

            if vars_values is None:
                return True

            p_value = vars_values.get(value.name)
            return self.is_assignable(p_value, type_registry)

        if isinstance(value, qm.NullValue):
            return True

        if not (isinstance(value, qm.ListValue) or isinstance(value, qm.ConstListValue)):
            return False

        of_type = self.of_type(type_registry)
        if not isinstance(of_type, _PossibleInputType):
            raise _SchemaAssertionError("Validating input for wrapper of non input type", "of_type")

        for i in value.values:
            if not of_type.validate_input(i, vars_values, vars_defs, type_registry):
                return False

        return True


class NonNull(_TypeWrapper):
    def _create_name(self, of_type: str) -> str:
        return of_type + '!'

    def is_assignable(self, value: PrimitiveType, type_registry: 'TypeRegistry') -> bool:
        if value is None:
            return False

        wrapped_type = self.of_type(type_registry)

        return wrapped_type.is_assignable(value, type_registry)

    def validate_input(self, value: t.Union[qm.Value, qm.ConstValue],
                       vars_values: t.Optional[t.Mapping[str, PrimitiveType]],
                       vars_defs: t.Mapping[str, GqlType],
                       type_registry: 'TypeRegistry') -> bool:
        if isinstance(value, qm.Variable):
            _validate_variable(value, self, vars_defs)

            if vars_values is None:
                return True

            p_value = vars_values.get(value.name)
            return self.is_assignable(p_value, type_registry)
        if isinstance(value, qm.NullValue):
            return False
        of_type = self.of_type(type_registry)
        if not isinstance(of_type, _PossibleInputType):
            raise _SchemaAssertionError("Validating input for wrapper of non input type", "of_type")
        return of_type.validate_input(value, vars_values, vars_defs, type_registry)


class Argument:
    def __init__(self, argument_type: t.Union['InlineType', str], default: 'PrimitiveType') -> None:
        self.__type = argument_type
        self.default = default
        self.__type_resolved: t.Optional[t.Union[InputType, WrapperType]] = None

    def is_assignable(self, value: PrimitiveType, type_registry: 'TypeRegistry') -> bool:
        self_type = self.type(type_registry)
        return self_type.is_assignable(value, type_registry)

    def validate_input(self, value: t.Union[qm.Value, qm.ConstValue],
                       vars_values: t.Optional[t.Mapping[str, PrimitiveType]],
                       vars_defs: t.Mapping[str, GqlType],
                       type_registry: 'TypeRegistry') -> bool:
        self_type = self.type(type_registry)
        return self_type.validate_input(value, vars_values, vars_defs, type_registry)

    def type(self, type_registry: 'TypeRegistry') -> t.Union['InputType', 'WrapperType']:
        if self.__type_resolved is None:
            self_type = type_registry.resolve_type(self.__type)
            self.__type_resolved = is_input(self_type)
            if self.__type_resolved is None:
                self.__type_resolved = assert_wrapper(self_type)
            assert_input(type_registry.resolve_and_unwrap(self.__type_resolved))
        return self.__type_resolved


class Field:
    def __init__(self, field_type: t.Union['InlineType', str], args: t.Mapping[str, Argument]) -> None:
        self.__type = field_type
        self.args = args
        self.__type_resolved: t.Optional[t.Union[OutputType, WrapperType]] = None

    def is_assignable(self, value: PrimitiveType, type_registry: 'TypeRegistry') -> bool:
        self_type = self.type(type_registry)
        return self_type.is_assignable(value, type_registry)

    def type(self, type_registry: 'TypeRegistry') -> t.Union['OutputType', 'WrapperType']:
        if self.__type_resolved is None:
            resolved_type = type_registry.resolve_type(self.__type)
            wrapper = is_wrapper(resolved_type)
            if wrapper is not None:
                self.__type_resolved = wrapper
            else:
                self.__type_resolved = assert_output(resolved_type)
        return self.__type_resolved


class _GqlCompositeType(GqlType):
    def __init__(self, name: str, fields: t.Mapping[str, Field]) -> None:
        super().__init__(name)
        self._fields = fields

    def fields(self, type_registry: 'TypeRegistry') -> t.Mapping[str, Field]:
        return self._fields

    def is_assignable(self, value: PrimitiveType, type_registry: 'TypeRegistry') -> bool:
        raise RuntimeError("Value must never be assigned to any composite type")


class _Scalar(_PossibleInputType):
    def __init__(self) -> None:
        super().__init__(type(self).__name__[1:])

    def is_assignable(self, value: PrimitiveType, type_registry: 'TypeRegistry') -> bool:
        raise NotImplementedError()

    def validate_input(self, value: t.Union[qm.Value, qm.ConstValue],
                       vars_values: t.Optional[t.Mapping[str, PrimitiveType]],
                       vars_defs: t.Mapping[str, GqlType],
                       type_registry: 'TypeRegistry') -> bool:
        raise NotImplementedError()


class _Boolean(_Scalar):
    def is_assignable(self, value: PrimitiveType, type_registry: 'TypeRegistry') -> bool:
        return value is None or isinstance(value, bool)

    def validate_input(self, value: t.Union[qm.Value, qm.ConstValue],
                       vars_values: t.Optional[t.Mapping[str, PrimitiveType]],
                       vars_defs: t.Mapping[str, GqlType],
                       type_registry: 'TypeRegistry') -> bool:
        if isinstance(value, qm.Variable):
            _validate_variable(value, self, vars_defs)

            if vars_values is None:
                return True

            p_value = vars_values.get(value.name)
            return self.is_assignable(p_value, type_registry)
        return isinstance(value, qm.BoolValue) or isinstance(value, qm.NullValue)


class _Int(_Scalar):
    def is_assignable(self, value: PrimitiveType, type_registry: 'TypeRegistry') -> bool:
        return value is None or isinstance(value, int)

    def validate_input(self, value: t.Union[qm.Value, qm.ConstValue],
                       vars_values: t.Optional[t.Mapping[str, PrimitiveType]],
                       vars_defs: t.Mapping[str, GqlType],
                       type_registry: 'TypeRegistry') -> bool:
        if isinstance(value, qm.Variable):
            _validate_variable(value, self, vars_defs)

            if vars_values is None:
                return True

            p_value = vars_values.get(value.name)
            return self.is_assignable(p_value, type_registry)
        return isinstance(value, qm.IntValue) or isinstance(value, qm.NullValue)


class _Float(_Scalar):
    def is_assignable(self, value: PrimitiveType, type_registry: 'TypeRegistry') -> bool:
        return value is None or isinstance(value, float)

    def validate_input(self, value: t.Union[qm.Value, qm.ConstValue],
                       vars_values: t.Optional[t.Mapping[str, PrimitiveType]],
                       vars_defs: t.Mapping[str, GqlType],
                       type_registry: 'TypeRegistry') -> bool:
        if isinstance(value, qm.Variable):
            _validate_variable(value, self, vars_defs)

            if vars_values is None:
                return True

            p_value = vars_values.get(value.name)
            return self.is_assignable(p_value, type_registry)
        return isinstance(value, qm.FloatValue) or isinstance(value, qm.NullValue)


class _String(_Scalar):
    def is_assignable(self, value: PrimitiveType, type_registry: 'TypeRegistry') -> bool:
        return value is None or isinstance(value, str)

    def validate_input(self, value: t.Union[qm.Value, qm.ConstValue],
                       vars_values: t.Optional[t.Mapping[str, PrimitiveType]],
                       vars_defs: t.Mapping[str, GqlType],
                       type_registry: 'TypeRegistry') -> bool:
        if isinstance(value, qm.Variable):
            _validate_variable(value, self, vars_defs)

            if vars_values is None:
                return True

            p_value = vars_values.get(value.name)
            return self.is_assignable(p_value, type_registry)
        return isinstance(value, qm.StrValue) or isinstance(value, qm.NullValue)


class _ID(_Scalar):
    def is_assignable(self, value: PrimitiveType, type_registry: 'TypeRegistry') -> bool:
        return value is None or isinstance(value, int) or isinstance(value, str)

    def validate_input(self, value: t.Union[qm.Value, qm.ConstValue],
                       vars_values: t.Optional[t.Mapping[str, PrimitiveType]],
                       vars_defs: t.Mapping[str, GqlType],
                       type_registry: 'TypeRegistry') -> bool:
        if isinstance(value, qm.Variable):
            _validate_variable(value, self, vars_defs)

            if vars_values is None:
                return True

            p_value = vars_values.get(value.name)
            return self.is_assignable(p_value, type_registry)
        return isinstance(value, qm.IntValue) or isinstance(value, qm.StrValue) or isinstance(value, qm.NullValue)


Boolean = _Boolean()
Int = _Int()
Float = _Float()
String = _String()
ID = _ID()


class Enum(GqlType):
    def __init__(self, name: str, values: t.Set[str]) -> None:
        super().__init__(name)
        if len(values) < 2:
            raise GqlSchemaError("Enum must define at least 2 possible values")
        self.values = values

    def is_assignable(self, value: PrimitiveType, type_registry: 'TypeRegistry') -> bool:
        return value is None or isinstance(value, str) and value in self.values

    def validate_input(self, value: t.Union[qm.Value, qm.ConstValue],
                       vars_values: t.Optional[t.Mapping[str, PrimitiveType]],
                       vars_defs: t.Mapping[str, GqlType],
                       type_registry: 'TypeRegistry') -> bool:
        if isinstance(value, qm.Variable):
            _validate_variable(value, self, vars_defs)

            if vars_values is None:
                return True

            p_value = vars_values.get(value.name)
            return self.is_assignable(p_value, type_registry)
        if isinstance(value, qm.NullValue):
            return True
        if not isinstance(value, qm.EnumValue):
            return False
        return value.value in self.values


class Interface(_GqlCompositeType):
    def of_objects(self, type_registry: 'TypeRegistry') -> t.Sequence['Object']:
        return type_registry.objects_by_interface(str(self))


class InputObject(_PossibleInputType):
    def __init__(self, name: str, fields: t.Mapping[str, t.Union['InlineType', str]]) -> None:
        super().__init__(name)

        if len(fields) == 0:
            raise GqlSchemaError("InputObject must define at least one field")

        self.__fields = fields
        self.__fields_resolved: t.Optional[t.Mapping[str, t.Union[InputType, WrapperType]]] = None

    def fields(self, type_registry: 'TypeRegistry') -> t.Mapping[str, t.Union['InputType', 'WrapperType']]:
        if self.__fields_resolved is None:
            fields_resolved: t.Dict[str, t.Union[InputType, WrapperType]] = {}
            for value_name, value_type in self.__fields.items():
                try:
                    value_type_resolved = type_registry.resolve_type(value_type)
                except TypeResolvingError as e:
                    raise _SchemaAssertionError(str(e), value_name) from e
                value_type_wrapper = is_wrapper(value_type_resolved)
                if value_type_wrapper is not None:
                    assert_input(type_registry.resolve_and_unwrap(value_type_wrapper))
                    fields_resolved[value_name] = value_type_wrapper
                else:
                    fields_resolved[value_name] = assert_input(value_type_resolved)
            self.__fields_resolved = fields_resolved
        return self.__fields_resolved

    def validate_input(self, value: t.Union[qm.Value, qm.ConstValue],
                       vars_values: t.Optional[t.Mapping[str, PrimitiveType]],
                       vars_defs: t.Mapping[str, GqlType],
                       type_registry: 'TypeRegistry') -> bool:
        if isinstance(value, qm.Variable):
            _validate_variable(value, self, vars_defs)

            if vars_values is None:
                return True

            p_value = vars_values.get(value.name)
            return self.is_assignable(p_value, type_registry)
        if isinstance(value, qm.NullValue):
            return True
        if isinstance(value, qm.ConstObjectValue) or isinstance(value, qm.ObjectValue):
            fields = self.fields(type_registry)

            value_keys = frozenset(value.values.keys())

            if len(value_keys.difference(fields)) > 0:
                return False

            for key, field_type in fields.items():
                field_value = value.values.get(key)
                if field_value is None:
                    field_value = qm.NullValue()
                if not field_type.validate_input(field_value, vars_values, vars_defs, type_registry):
                    return False
            return True
        return False

    def is_assignable(self, value: PrimitiveType, type_registry: 'TypeRegistry') -> bool:
        if value is None:
            return True

        if not isinstance(value, dict):
            return False

        value_keys = frozenset(value.keys())
        fields = self.fields(type_registry)

        if len(value_keys.difference(fields.keys())) > 0:
            return False

        for value_name, field_type in fields.items():
            if not field_type.is_assignable(value.get(value_name), type_registry):
                return False

        return True


class Object(_GqlCompositeType):
    def __init__(self, name: str, fields: t.Mapping[str, Field],
                 implements: t.Set[str]) -> None:
        super().__init__(name, fields)

        if len(fields) == 0 and len(implements) == 0:
            raise GqlSchemaError("Object must define at least one field or implement interface")

        self.__implements = implements
        self.__implements_resolved: t.Optional[t.Sequence[Interface]] = None

    def fields(self, type_registry: 'TypeRegistry') -> t.Mapping[str, Field]:
        fields: t.Dict[str, Field] = {}
        for i in self.implements(type_registry):
            for n, f in i.fields(type_registry).items():
                fields[n] = f
        for n, f in self._fields.items():
            fields[n] = f
        return fields

    def own_fields(self) -> t.Mapping[str, Field]:
        return self._fields

    def implements(self, type_registry: 'TypeRegistry') -> t.Sequence[Interface]:
        if self.__implements_resolved is None:
            self.__implements_resolved = []
            for interface_name in self.__implements:
                interface = type_registry.resolve_type(interface_name)
                if not isinstance(interface, Interface):
                    raise _SchemaAssertionError("Interface expected here", interface_name)
                self.__implements_resolved.append(interface)
        return self.__implements_resolved


class Union(GqlType):
    def __init__(self, name: str, of_objects: t.Set[str]) -> None:
        super().__init__(name)

        if len(of_objects) < 2:
            raise GqlSchemaError("Union must unite at least 2 objects")

        self.__of_objects = of_objects
        self.__of_objects_resolved: t.Optional[t.Sequence[Object]] = None

    def of_objects(self, type_registry: 'TypeRegistry') -> t.Sequence[Object]:
        if self.__of_objects_resolved is None:
            self.__of_objects_resolved = []
            for object_name in self.__of_objects:
                object_type = type_registry.resolve_type(object_name)
                if not isinstance(object_type, Object):
                    raise _SchemaAssertionError("Object expected here", object_name)
                self.__of_objects_resolved.append(object_type)
        return self.__of_objects_resolved

    def is_assignable(self, value: PrimitiveType, type_registry: 'TypeRegistry') -> bool:
        raise RuntimeError("Value must never be assigned to union")


ScalarType = t.Union[_Boolean, _Int, _Float, _String, _ID]


def is_scalar(type_to_test: GqlType) -> t.Optional[ScalarType]:
    if isinstance(type_to_test, _Boolean):
        return type_to_test
    if isinstance(type_to_test, _Int):
        return type_to_test
    if isinstance(type_to_test, _Float):
        return type_to_test
    if isinstance(type_to_test, _String):
        return type_to_test
    if isinstance(type_to_test, _ID):
        return type_to_test
    return None


def assert_scalar(type_to_test: GqlType) -> ScalarType:
    scalar = is_scalar(type_to_test)
    if scalar is None:
        raise _SchemaAssertionError("Scalar expected here", str(type_to_test))
    return scalar


WrapperType = t.Union[List, NonNull]


def is_wrapper(type_to_test: GqlType) -> t.Optional[WrapperType]:
    if isinstance(type_to_test, List):
        return type_to_test
    if isinstance(type_to_test, NonNull):
        return type_to_test
    return None


def assert_wrapper(type_to_test: GqlType) -> WrapperType:
    wrapper = is_wrapper(type_to_test)
    if wrapper is None:
        raise _SchemaAssertionError("Wrapper expected here", str(type_to_test))
    return wrapper


NonWrapperType = t.Union[ScalarType, Enum, Interface, Object, Union, InputObject]


def is_non_wrapper(type_to_test: GqlType) -> t.Optional[NonWrapperType]:
    if is_wrapper(type_to_test) is not None:
        return None
    return t.cast(NonWrapperType, type_to_test)


def assert_non_wrapper(type_to_test: GqlType) -> NonWrapperType:
    non_wrapper = is_non_wrapper(type_to_test)
    if non_wrapper is None:
        raise _SchemaAssertionError("Non wrapper expected here", str(type_to_test))
    return non_wrapper


SpreadableType = t.Union[Interface, Object, Union]


def is_spreadable(type_to_test: GqlType) -> t.Optional[SpreadableType]:
    if isinstance(type_to_test, Interface):
        return type_to_test
    if isinstance(type_to_test, Object):
        return type_to_test
    if isinstance(type_to_test, Union):
        return type_to_test
    return None


def assert_spreadable(type_to_test: GqlType) -> SpreadableType:
    spreadable = is_spreadable(type_to_test)
    if spreadable is None:
        raise _SchemaAssertionError("Spreadable expected here", str(type_to_test))
    return spreadable


SelectableType = t.Union[Interface, Object]


def is_selectable(type_to_test: GqlType) -> t.Optional[SelectableType]:
    if isinstance(type_to_test, Interface):
        return type_to_test
    if isinstance(type_to_test, Object):
        return type_to_test
    return None


def assert_selectable(type_to_test: GqlType) -> SelectableType:
    selectable = is_selectable(type_to_test)
    if selectable is None:
        raise _SchemaAssertionError("Selectable expected here", str(type_to_test))
    return selectable


InputType = t.Union[ScalarType, Enum, InputObject]


def is_input(type_to_test: GqlType) -> t.Optional[InputType]:
    if isinstance(type_to_test, _Boolean):
        return type_to_test
    if isinstance(type_to_test, _Int):
        return type_to_test
    if isinstance(type_to_test, _Float):
        return type_to_test
    if isinstance(type_to_test, _String):
        return type_to_test
    if isinstance(type_to_test, _ID):
        return type_to_test
    if isinstance(type_to_test, Enum):
        return type_to_test
    if isinstance(type_to_test, InputObject):
        return type_to_test
    return None


def assert_input(type_to_test: GqlType) -> InputType:
    input_type = is_input(type_to_test)
    if input_type is None:
        raise _SchemaAssertionError("Input type expected here", str(type_to_test))
    return input_type


OutputType = t.Union[ScalarType, Enum, Interface, Object, Union]


def is_output(type_to_test: GqlType) -> t.Optional[OutputType]:
    if isinstance(type_to_test, InputObject):
        return None
    if is_wrapper(type_to_test) is not None:
        return None
    return t.cast(OutputType, type_to_test)


def assert_output(type_to_test: GqlType) -> OutputType:
    output = is_output(type_to_test)
    if output is None:
        raise _SchemaAssertionError("Output type expected here", str(type_to_test))
    return output


UserType = t.Union[Enum, Interface, Object, Union, InputObject]


def is_user(type_to_test: GqlType) -> t.Optional[UserType]:
    if isinstance(type_to_test, Enum):
        return type_to_test
    if isinstance(type_to_test, Interface):
        return type_to_test
    if isinstance(type_to_test, Object):
        return type_to_test
    if isinstance(type_to_test, Union):
        return type_to_test
    if isinstance(type_to_test, InputObject):
        return type_to_test
    return None


def assert_user(type_to_test: GqlType) -> UserType:
    user = is_user(type_to_test)
    if user is None:
        raise _SchemaAssertionError("User defined type expected here", str(type_to_test))
    return user


InlineType = t.Union['ScalarType', 'WrapperType']


def is_inline(type_to_test: GqlType) -> t.Optional[t.Union['ScalarType', 'WrapperType']]:
    scalar = is_scalar(type_to_test)
    if scalar is not None:
        return scalar
    wrapper = is_wrapper(type_to_test)
    if wrapper is not None:
        return wrapper
    return None


def assert_inline(type_to_test: GqlType) -> t.Union['ScalarType', 'WrapperType']:
    inline = is_inline(type_to_test)
    if inline is None:
        raise _SchemaAssertionError("Inline type expected expected here", str(type_to_test))
    return inline


DirectiveLocation = t.NewType("DirectiveLocation", str)


class DirectiveLocations:
    QUERY = DirectiveLocation("QUERY")
    MUTATION = DirectiveLocation("MUTATION")
    FIELD = DirectiveLocation("FIELD")
    FRAGMENT_DEFINITION = DirectiveLocation("FRAGMENT_DEFINITION")
    FRAGMENT_SPREAD = DirectiveLocation("FRAGMENT_SPREAD")
    INLINE_FRAGMENT = DirectiveLocation("INLINE_FRAGMENT")


class Directive:
    def __init__(self, name: str, locations: t.Set[DirectiveLocation], args: t.Mapping[str, Argument]) -> None:
        self.__name = name
        self.locations = locations
        self.args = args

    def __str__(self) -> str:
        return "@" + self.__name


class TypeRegistry:
    def __init__(self, type_defs: t.Sequence[UserType], directives: t.Sequence[Directive]) -> None:
        for user_type in type_defs:
            if str(user_type) in {"Boolean", "Int", "Float", "String", "ID",
                                  "__Schema", "__Type", "__Field", "__InputValue", "__EnumValue", "__TypeKind",
                                  "__Directive", "__DirectiveLocation"}:
                raise GqlSchemaError("Can not redeclare standard type `{}`".format(str(user_type)))

        self.__types: t.List[GqlType] = list(type_defs)

        self.__types.extend((
            Boolean,
            Int,
            Float,
            String,
            ID,

            Object("__Schema", {
                "types": Field(NonNull(List(NonNull("__Type"))), {}),
                "queryType": Field(NonNull("__Type"), {}),
                "mutationType": Field("__Type", {}),
                "directives": Field(NonNull(List(NonNull("__Directive"))), {})
            }, set()),

            Object("__Type", {
                "kind": Field(NonNull("__TypeKind"), {}),
                "name": Field(String, {}),
                "description": Field(String, {}),
                "fields": Field(
                    List(NonNull("__Field")),
                    {"includeDeprecated": Argument(Boolean, False)}
                ),
                "interfaces": Field(List(NonNull("__Type")), {}),
                "possibleTypes": Field(List(NonNull("__Type")), {}),
                "enumValues": Field(
                    List(NonNull("__EnumValue")),
                    {"includeDeprecated": Argument(Boolean, False)}
                ),
                "inputFields": Field(List(NonNull("__InputValue")), {}),
                "ofType": Field("__Type", {})
            }, set()),

            Object("__Field", {
                "name": Field(NonNull(String), {}),
                "description": Field(String, {}),
                "args": Field(NonNull(List(NonNull("__InputValue"))), {}),
                "type": Field(NonNull("__Type"), {}),
                "isDeprecated": Field(NonNull(Boolean), {}),
                "deprecationReason": Field(String, {})
            }, set()),

            Object("__InputValue", {
                "name": Field(NonNull(String), {}),
                "description": Field(String, {}),
                "type": Field(NonNull("__Type"), {}),
                "defaultValue": Field(String, {})
            }, set()),

            Object("__EnumValue", {
                "name": Field(NonNull(String), {}),
                "description": Field(String, {}),
                "isDeprecated": Field(NonNull(Boolean), {}),
                "deprecationReason": Field(String, {})
            }, set()),

            Enum("__TypeKind", {
                "SCALAR",
                "OBJECT",
                "INTERFACE",
                "UNION",
                "ENUM",
                "INPUT_OBJECT",
                "LIST",
                "NON_NULL"
            }),

            Object("__Directive", {
                "name": Field(NonNull(String), {}),
                "description": Field(String, {}),
                "locations": Field(NonNull(List(NonNull("__DirectiveLocation"))), {}),
                "args": Field(NonNull(List(NonNull("__InputValue"))), {})
            }, set()),

            Enum("__DirectiveLocation", {
                "QUERY",
                "MUTATION",
                "FIELD",
                "FRAGMENT_DEFINITION",
                "FRAGMENT_SPREAD",
                "INLINE_FRAGMENT"
            })
        ))
        self.__types_by_names: t.Dict[str, GqlType] = {}

        for type_def in self.__types:
            if str(type_def) in self.__types_by_names:
                raise GqlSchemaError("Type re-definition; problem with `{}` type".format(str(type_def)))
            self.__types_by_names[str(type_def)] = type_def

        self.__directives = list(directives)

        self.__directives.append(Directive(
            "skip",
            {DirectiveLocations.INLINE_FRAGMENT, DirectiveLocations.FRAGMENT_SPREAD, DirectiveLocations.FIELD},
            {"if": Argument(Boolean, None)}
        ))
        self.__directives.append(Directive(
            "include",
            {DirectiveLocations.INLINE_FRAGMENT, DirectiveLocations.FRAGMENT_SPREAD, DirectiveLocations.FIELD},
            {"if": Argument(Boolean, None)}
        ))

        self.__directives_by_names = dict(((str(directive)[1:], directive) for directive in self.__directives))

        self.__validate()

        self.__objects_by_interfaces: t.Dict[str, t.List[Object]] = {}

        for gql_type in self.__types:
            if isinstance(gql_type, Object):
                for interface in gql_type.implements(self):
                    self.__objects_by_interfaces.setdefault(str(interface), [])
                    objects = self.__objects_by_interfaces[str(interface)]
                    objects.append(gql_type)
            if isinstance(gql_type, Interface):
                self.__objects_by_interfaces.setdefault(str(gql_type), [])

    def resolve_type(self, type_or_ref: t.Union[GqlType, str]) -> GqlType:
        if isinstance(type_or_ref, GqlType):
            return type_or_ref
        if type_or_ref not in self.__types_by_names:
            raise TypeResolvingError(type_or_ref)
        return self.__types_by_names[type_or_ref]

    def resolve_and_unwrap(self, type_or_ref: t.Union[GqlType, str]) -> NonWrapperType:
        gql_type = self.resolve_type(type_or_ref)

        wrapper = is_wrapper(gql_type)
        while wrapper is not None:
            gql_type = wrapper.of_type(self)
            wrapper = is_wrapper(gql_type)

        return assert_non_wrapper(gql_type)

    def objects_by_interface(self, interface_name: str) -> t.Sequence[Object]:
        return self.__objects_by_interfaces[interface_name]

    def directive(self, name: str) -> Directive:
        if name in self.__directives_by_names:
            return self.__directives_by_names[name]

        raise TypeResolvingError("Directive `{}` not found".format(name))

    __NAME_RE = re.compile(r'^[_A-Za-z][_0-9A-Za-z]*$')

    def __validate(self) -> None:
        for user_type in self.__types:
            if self.__NAME_RE.match(str(user_type)) is None:
                raise GqlSchemaError(
                    "Wrong type name: /{}/ expected, but got '{}'".format(self.__NAME_RE.pattern[1:-1], str(user_type)))

            if isinstance(user_type, Interface):
                self.__validate_fields(user_type)
            if isinstance(user_type, Object):
                self.__validate_object(user_type)
            if isinstance(user_type, Union):
                try:
                    user_type.of_objects(self)
                except TypeResolvingError as e:
                    raise GqlSchemaError(
                        "{}; problem with `{}` union".format(str(e), str(user_type))
                    ) from e
                except _SchemaAssertionError as e:
                    raise GqlSchemaError(
                        "Union must unite only objects; "
                        "problem with `{}` object of `{}` union".format(e.args[1], str(user_type))
                    ) from e
            if isinstance(user_type, InputObject):
                try:
                    declared_fields = user_type.fields(self)
                except _SchemaAssertionError as e:
                    raise GqlSchemaError(
                        "{}; problem with `{}` field of `{}` input object".format(str(e), e.args[1], str(user_type))
                    ) from e
                for field_name in declared_fields.keys():
                    if self.__NAME_RE.match(str(field_name)) is None:
                        raise GqlSchemaError(
                            "Wrong name of field: /{}/ expected, but got '{}' in `{}` type".format(
                                self.__NAME_RE.pattern[1:-1], field_name, str(user_type)
                            )
                        )
        for directive in self.__directives:
            if self.__NAME_RE.match(str(directive)[1:]) is None:
                raise GqlSchemaError(
                    "Wrong name of directive: /{}/ expected, but got '{}'".format(
                        self.__NAME_RE.pattern[1:-1], str(directive)[1:]
                    ))
            try:
                self.__validate_args(directive.args)
            except _SchemaAssertionError as e:
                raise GqlSchemaError(
                    "{}; problem with `{}` argument of `{}` directive".format(
                        str(e), e.args[1], str(directive)[1:]
                    )
                )

    def __validate_fields(self, selectable_type: SelectableType) -> None:
        for field_name, field in selectable_type.fields(self).items():
            if self.__NAME_RE.match(str(field_name)) is None:
                raise GqlSchemaError(
                    "Wrong name of field: /{}/ expected, but got '{}' in `{}` type".format(
                        self.__NAME_RE.pattern[1:-1], field_name, str(selectable_type)
                    )
                )

            try:
                field.type(self)
            except TypeResolvingError as e:
                raise GqlSchemaError(
                    "{}; problem with `{}` field of `{}` type".format(str(e), field_name, str(selectable_type))
                )
            except _SchemaAssertionError as e:
                raise GqlSchemaError(
                    "{}; problem with `{}` field of `{}` type".format(
                        str(e), field_name, str(selectable_type)
                    )
                ) from e
            try:
                self.__validate_args(field.args)
            except _SchemaAssertionError as e:
                raise GqlSchemaError(
                    "{}; problem with `{}` argument of `{}` field of `{}` type".format(
                        str(e), e.args[1], field_name, str(selectable_type)
                    )
                ) from e

    def __validate_args(self, args: t.Mapping[str, Argument]) -> None:
        for arg_name, arg in args.items():
            if self.__NAME_RE.match(str(arg_name)) is None:
                raise _SchemaAssertionError(
                    "Wrong name of argument: /{}/ expected, but got '{}'".format(
                        self.__NAME_RE.pattern[1:-1], arg_name
                    ), arg_name
                )

            try:
                arg_type = self.resolve_type(arg.type(self))
            except TypeResolvingError as e:
                raise _SchemaAssertionError(str(e), arg_name) from e
            except _SchemaAssertionError as e:
                raise _SchemaAssertionError("Input type expected", arg_name) from e
            if arg.default is not None and not arg_type.is_assignable(arg.default, self):
                raise _SchemaAssertionError("{} is not assignable to `{}` type".format(
                    json.dumps(arg.default), str(arg_type)
                ), arg_name)

    def __validate_object(self, obj: Object) -> None:
        try:
            interfaces = list(obj.implements(self))
        except TypeResolvingError as e:
            raise GqlSchemaError(
                "{}; problem with `{}` object".format(str(e), str(obj))
            ) from e
        except _SchemaAssertionError as e:
            raise GqlSchemaError(
                "Object must implement only interfaces; "
                "problem with `{}` interface of `{}` object".format(e.args[1], str(obj))
            ) from e
        declared_fields: t.Dict[str, str] = {}
        interfaces.sort(key=lambda i: str(i))
        for interface in interfaces:
            for field_name in interface.fields(self).keys():
                if field_name in declared_fields:
                    raise GqlSchemaError(
                        "Interfaces `{}` and `{}` of `{}` object both declare `{}` field".format(
                            declared_fields[field_name], str(interface), str(obj), field_name
                        )
                    )
                declared_fields[field_name] = str(interface)
        for field_name in obj.own_fields().keys():
            if field_name in declared_fields:
                raise GqlSchemaError(
                    "Object `{}` redeclare `{}` field of `{}` interface".format(
                        str(obj), field_name, declared_fields[field_name]
                    )
                )
        self.__validate_fields(obj)


def _validate_variable(var: qm.Variable, expected: GqlType, vars_defs: t.Mapping[str, GqlType]) -> None:
    if var.name not in vars_defs:
        raise UndefinedVariableError(var.name)

    var_def = vars_defs[var.name]
    if str(expected) != str(var_def):
        raise NonCompatibleVariableType(var.name, expected, var_def)


__all__ = ["TypeResolvingError", "GqlType", "List", "NonNull", "Argument", "Field", "Boolean", "Int", "Float",
           "String", "ID", "Enum", "Interface", "InputObject", "Object", "Union", "ScalarType", "is_scalar",
           "assert_scalar", "WrapperType", "is_wrapper", "assert_wrapper", "NonWrapperType", "is_non_wrapper",
           "assert_non_wrapper", "SpreadableType", "is_spreadable", "assert_spreadable", "SelectableType",
           "is_selectable", "assert_selectable", "InputType", "is_input", "assert_input", "OutputType",
           "is_output", "assert_output", "UserType", "is_user", "assert_user", "InlineType", "is_inline",
           "assert_inline", "DirectiveLocation", "DirectiveLocations", "Directive", "TypeRegistry"]



import typing as t

PrimitiveType = t.Union[None, bool, int, float, str, t.Sequence['PrimitiveType'], t.Mapping[str, 'PrimitiveType']]


class PrimitiveSerializable:
    def to_primitive(self) -> PrimitiveType:
        raise NotImplementedError()


def add_if_not_empty(mapping: t.MutableMapping[str, PrimitiveType],
                     name: str, values: t.Optional[t.Sequence[PrimitiveSerializable]]) -> None:
    if values is not None and len(values) > 0:
        mapping[name] = [i.to_primitive() for i in values]


def add_if_not_none(mapping: t.MutableMapping[str, PrimitiveType], name: str, value: PrimitiveType) -> None:
    if value is not None:
        mapping[name] = value


__all__ = ["PrimitiveType", "PrimitiveSerializable", "add_if_not_empty", "add_if_not_none"]



from typing import Sequence, Union, Optional, Mapping, Dict, Tuple

from .utils import add_if_not_empty, add_if_not_none, PrimitiveType, PrimitiveSerializable


class QueryVisitor:
    def visit_document_begin(self, document: 'Document') -> None:
        pass

    def visit_document_end(self, document: 'Document') -> None:
        pass

    def visit_query_begin(self, query: 'Query') -> None:
        pass

    def visit_query_end(self, query: 'Query') -> None:
        pass

    def visit_mutation_begin(self, mutation: 'Mutation') -> None:
        pass

    def visit_mutation_end(self, mutation: 'Mutation') -> None:
        pass

    def visit_variable_definition(self, var_def: 'VariableDefinition') -> None:
        pass

    def visit_directive_begin(self, directive: 'Directive') -> None:
        pass

    def visit_directive_end(self, directive: 'Directive') -> None:
        pass

    def visit_argument(self, argument: 'Argument') -> None:
        pass

    def visit_fragment_begin(self, fragment: 'NamedFragment') -> None:
        pass

    def visit_fragment_end(self, fragment: 'NamedFragment') -> None:
        pass

    def visit_field_selection_begin(self, field_sel: 'FieldSelection') -> None:
        pass

    def visit_field_selection_end(self, field_sel: 'FieldSelection') -> None:
        pass

    def visit_fragment_spread_begin(self, spread: 'FragmentSpread') -> None:
        pass

    def visit_fragment_spread_end(self, spread: 'FragmentSpread') -> None:
        pass

    def visit_inline_fragment_begin(self, fragment: 'InlineFragment') -> None:
        pass

    def visit_inline_fragment_end(self, fragment: 'InlineFragment') -> None:
        pass

    def visit_variable(self, var: 'Variable') -> None:
        pass

    def visit_null_value(self, val: 'NullValue') -> None:
        pass

    def visit_bool_value(self, val: 'BoolValue') -> None:
        pass

    def visit_int_value(self, val: 'IntValue') -> None:
        pass

    def visit_float_value(self, val: 'FloatValue') -> None:
        pass

    def visit_str_value(self, val: 'StrValue') -> None:
        pass

    def visit_enum_value(self, val: 'EnumValue') -> None:
        pass

    def visit_list_value_begin(self, val: Union['ListValue', 'ConstListValue']) -> None:
        pass

    def visit_list_value_end(self, val: Union['ListValue', 'ConstListValue']) -> None:
        pass

    def visit_object_value_begin(self, val: Union['ObjectValue', 'ConstObjectValue']) -> None:
        pass

    def visit_object_field(self, val: Tuple[str, Union['Value', 'ConstValue']]) -> None:
        pass

    def visit_object_value_end(self, val: Union['ObjectValue', 'ConstObjectValue']) -> None:
        pass


class GraphQlModelType(PrimitiveSerializable):
    def visit(self, visitor: QueryVisitor) -> None:
        raise NotImplementedError()

    def to_primitive(self) -> PrimitiveType:
        raise NotImplementedError()


class Document(GraphQlModelType):
    operations: Sequence['Operation']
    fragments: Sequence['NamedFragment']

    def __init__(self, operations: Sequence['Operation'], fragments: Sequence['NamedFragment']) -> None:
        self.operations = operations
        self.fragments = fragments

    def visit(self, visitor: QueryVisitor) -> None:
        visitor.visit_document_begin(self)

        for op in self.operations:
            op.visit(visitor)

        for fr in self.fragments:
            fr.visit(visitor)

        visitor.visit_document_end(self)

    def to_primitive(self) -> PrimitiveType:
        p: Dict[str, PrimitiveType] = {"@doc": None}
        add_if_not_empty(p, "operations", self.operations)
        add_if_not_empty(p, "fragments", self.fragments)
        return p


class Operation(GraphQlModelType):
    name: Optional[str]
    variables: Sequence['VariableDefinition']
    directives: Sequence['Directive']
    selections: Sequence['Selection']

    def __init__(self,
                 name: Optional[str],
                 variables: Sequence['VariableDefinition'],
                 directives: Sequence['Directive'],
                 selections: Sequence['Selection']) -> None:
        self.name = name
        self.variables = variables
        self.directives = directives
        self.selections = selections

    def to_primitive(self) -> PrimitiveType:
        p: Dict[str, PrimitiveType] = {"@m": self.name}

        if type(self).__name__ == "Query":
            p = {"@q": self.name}

        add_if_not_empty(p, "variables", self.variables)
        add_if_not_empty(p, "directives", self.directives)
        add_if_not_empty(p, "selections", self.selections)
        return p

    def visit(self, visitor: QueryVisitor) -> None:
        raise NotImplementedError()

    def _visit_internal(self, visitor: QueryVisitor) -> None:
        for v_def in self.variables:
            v_def.visit(visitor)

        for directive in self.directives:
            directive.visit(visitor)

        for sel in self.selections:
            sel.visit(visitor)


class Query(Operation):
    def visit(self, visitor: QueryVisitor) -> None:
        visitor.visit_query_begin(self)

        self._visit_internal(visitor)

        visitor.visit_query_end(self)


class Mutation(Operation):
    def visit(self, visitor: QueryVisitor) -> None:
        visitor.visit_mutation_begin(self)

        self._visit_internal(visitor)

        visitor.visit_mutation_end(self)


class VariableDefinition(GraphQlModelType):
    name: str
    type: 'Type'
    default: Optional['ConstValue']

    def __init__(self, name: str, var_type: 'Type', default: Optional['ConstValue']) -> None:
        self.name = name
        self.type = var_type
        self.default = default

    def visit(self, visitor: QueryVisitor) -> None:
        visitor.visit_variable_definition(self)

    def to_primitive(self) -> PrimitiveType:
        return [self.name, self.type.to_primitive(), None if self.default is None else self.default.to_primitive()]


class Type(GraphQlModelType):
    null: bool

    def __init__(self, null: bool) -> None:
        self.null = null

    def visit(self, visitor: QueryVisitor) -> None:
        raise RuntimeError("Type is not suppose to be visited")

    def to_primitive(self) -> PrimitiveType:
        raise NotImplementedError()


class NamedType(Type):
    name: str

    def __init__(self, name: str, null: bool) -> None:
        super().__init__(null)
        self.name = name

    def to_primitive(self) -> PrimitiveType:
        key = "@named"

        if not self.null:
            key += "!"

        return {key: self.name}


class ListType(Type):
    el_type: Type

    def __init__(self, el_type: Type, null: bool) -> None:
        super().__init__(null)
        self.el_type = el_type

    def to_primitive(self) -> PrimitiveType:
        key = "@list"

        if not self.null:
            key += "!"

        return {key: self.el_type.to_primitive()}


class Directive(GraphQlModelType):
    name: str
    arguments: Sequence['Argument']

    def __init__(self, name: str, arguments: Sequence['Argument']) -> None:
        self.name = name
        self.arguments = arguments

    def visit(self, visitor: QueryVisitor) -> None:
        visitor.visit_directive_begin(self)

        for arg in self.arguments:
            arg.visit(visitor)

        visitor.visit_directive_end(self)

    def to_primitive(self) -> PrimitiveType:
        d: Dict[str, PrimitiveType] = {"@dir": self.name}
        add_if_not_empty(d, "arguments", self.arguments)
        return d


ConstValue = Union['IntValue', 'FloatValue', 'StrValue', 'BoolValue', 'NullValue', 'EnumValue', 'ConstListValue',
                   'ConstObjectValue']

Value = Union['Variable', 'IntValue', 'FloatValue', 'StrValue', 'BoolValue', 'NullValue', 'EnumValue', 'ListValue',
              'ObjectValue']


class ValueModelType(GraphQlModelType):
    def visit(self, visitor: QueryVisitor) -> None:
        raise NotImplementedError()

    def to_primitive(self) -> PrimitiveType:
        raise NotImplementedError()

    def to_py_value(self, variables: Mapping[str, PrimitiveType]) -> PrimitiveType:
        raise NotImplementedError()


class Variable(ValueModelType):
    name: str

    def __init__(self, name: str) -> None:
        self.name = name

    def visit(self, visitor: QueryVisitor) -> None:
        visitor.visit_variable(self)

    def to_primitive(self) -> PrimitiveType:
        return {"@var": self.name}

    def to_py_value(self, variables: Mapping[str, PrimitiveType]) -> PrimitiveType:
        return variables.get(self.name)


class NullValue(ValueModelType):
    def to_primitive(self) -> PrimitiveType:
        return {"@null": None}

    def visit(self, visitor: QueryVisitor) -> None:
        visitor.visit_null_value(self)

    def to_py_value(self, variables: Mapping[str, PrimitiveType]) -> PrimitiveType:
        return None


class EnumValue(ValueModelType):
    value: str

    def __init__(self, value: str) -> None:
        self.value = value

    def visit(self, visitor: QueryVisitor) -> None:
        visitor.visit_enum_value(self)

    def to_primitive(self) -> PrimitiveType:
        return {"@enum": self.value}

    def to_py_value(self, variables: Mapping[str, PrimitiveType]) -> PrimitiveType:
        return self.value


class IntValue(ValueModelType):
    value: int

    def __init__(self, value: int) -> None:
        self.value = value

    def visit(self, visitor: QueryVisitor) -> None:
        visitor.visit_int_value(self)

    def to_primitive(self) -> PrimitiveType:
        return {"@int": self.value}

    def to_py_value(self, variables: Mapping[str, PrimitiveType]) -> PrimitiveType:
        return self.value


class FloatValue(ValueModelType):
    value: float

    def __init__(self, value: float) -> None:
        self.value = value

    def visit(self, visitor: QueryVisitor) -> None:
        visitor.visit_float_value(self)

    def to_primitive(self) -> PrimitiveType:
        return {"@float": self.value}

    def to_py_value(self, variables: Mapping[str, PrimitiveType]) -> PrimitiveType:
        return self.value


class StrValue(ValueModelType):
    value: str

    def __init__(self, value: str) -> None:
        self.value = value

    def visit(self, visitor: QueryVisitor) -> None:
        visitor.visit_str_value(self)

    def to_primitive(self) -> PrimitiveType:
        return {"@str": self.value}

    def to_py_value(self, variables: Mapping[str, PrimitiveType]) -> PrimitiveType:
        return self.value


class BoolValue(ValueModelType):
    value: bool

    def __init__(self, value: bool) -> None:
        self.value = value

    def visit(self, visitor: QueryVisitor) -> None:
        visitor.visit_bool_value(self)

    def to_primitive(self) -> PrimitiveType:
        return {"@bool": self.value}

    def to_py_value(self, variables: Mapping[str, PrimitiveType]) -> PrimitiveType:
        return self.value


class ConstListValue(ValueModelType):
    values: Sequence[ConstValue]

    def __init__(self, values: Sequence[ConstValue]) -> None:
        self.values = values

    def visit(self, visitor: QueryVisitor) -> None:
        visitor.visit_list_value_begin(self)

        for val in self.values:
            val.visit(visitor)

        visitor.visit_list_value_end(self)

    def to_primitive(self) -> PrimitiveType:
        return {"@const-list": [v.to_primitive() for v in self.values]}

    def to_py_value(self, variables: Mapping[str, PrimitiveType]) -> PrimitiveType:
        return [value.to_py_value(variables) for value in self.values]


class ListValue(ValueModelType):
    values: Sequence[Value]

    def __init__(self, values: Sequence[Value]) -> None:
        self.values = values

    def visit(self, visitor: QueryVisitor) -> None:
        visitor.visit_list_value_begin(self)

        for val in self.values:
            val.visit(visitor)

        visitor.visit_list_value_end(self)

    def to_primitive(self) -> PrimitiveType:
        return {"@list": [v.to_primitive() for v in self.values]}

    def to_py_value(self, variables: Mapping[str, PrimitiveType]) -> PrimitiveType:
        return [value.to_py_value(variables) for value in self.values]


class ConstObjectValue(ValueModelType):
    values: Mapping[str, ConstValue]

    def __init__(self, values: Mapping[str, ConstValue]) -> None:
        self.values = values

    def visit(self, visitor: QueryVisitor) -> None:
        visitor.visit_object_value_begin(self)

        for field in self.values.items():
            visitor.visit_object_field(field)
            field[1].visit(visitor)

        visitor.visit_object_value_end(self)

    def to_primitive(self) -> PrimitiveType:
        return {"@const-obj": dict(((k, v.to_primitive()) for k, v in self.values.items()))}

    def to_py_value(self, variables: Mapping[str, PrimitiveType]) -> PrimitiveType:
        return dict(((name, value.to_py_value(variables)) for name, value in self.values.items()))


class ObjectValue(ValueModelType):
    values: Mapping[str, Value]

    def __init__(self, values: Mapping[str, Value]) -> None:
        self.values = values

    def visit(self, visitor: QueryVisitor) -> None:
        visitor.visit_object_value_begin(self)

        for field in self.values.items():
            visitor.visit_object_field(field)
            field[1].visit(visitor)

        visitor.visit_object_value_end(self)

    def to_primitive(self) -> PrimitiveType:
        return {"@obj": dict(((k, v.to_primitive()) for k, v in self.values.items()))}

    def to_py_value(self, variables: Mapping[str, PrimitiveType]) -> PrimitiveType:
        return dict(((name, value.to_py_value(variables)) for name, value in self.values.items()))


class Argument(GraphQlModelType):
    name: str
    value: Value

    def __init__(self, name: str, value: Value) -> None:
        self.name = name
        self.value = value

    def visit(self, visitor: QueryVisitor) -> None:
        visitor.visit_argument(self)
        self.value.visit(visitor)

    def to_primitive(self) -> PrimitiveType:
        return [self.name, self.value.to_primitive()]


class Fragment(GraphQlModelType):
    on_type: Optional[NamedType]
    directives: Sequence[Directive]
    selections: Sequence['Selection']

    def visit(self, visitor: QueryVisitor) -> None:
        raise NotImplementedError()

    def to_primitive(self) -> PrimitiveType:
        raise NotImplementedError()


class NamedFragment(Fragment):
    name: str
    on_type: NamedType
    directives: Sequence[Directive]
    selections: Sequence['Selection']

    def __init__(self, name: str, on_type: NamedType,
                 directives: Sequence[Directive], selections: Sequence['Selection']) -> None:
        self.name = name
        self.on_type = on_type
        self.directives = directives
        self.selections = selections

    def visit(self, visitor: QueryVisitor) -> None:
        visitor.visit_fragment_begin(self)

        for directive in self.directives:
            directive.visit(visitor)

        for sel in self.selections:
            sel.visit(visitor)

        visitor.visit_fragment_end(self)

    def to_primitive(self) -> PrimitiveType:
        d = {"@frg": self.name, "on_type": self.on_type.to_primitive()}

        add_if_not_empty(d, "directives", self.directives)
        add_if_not_empty(d, "selections", self.selections)
        return d


class Selection(GraphQlModelType):
    def visit(self, visitor: QueryVisitor) -> None:
        raise NotImplementedError()

    def to_primitive(self) -> PrimitiveType:
        raise NotImplementedError()


class FieldSelection(Selection):
    alias: Optional[str]
    name: str
    arguments: Sequence[Argument]
    directives: Sequence[Directive]
    selections: Sequence[Selection]

    def __init__(self, alias: Optional[str],
                 name: str,
                 arguments: Sequence[Argument],
                 directives: Sequence[Directive],
                 selections: Sequence[Selection]) -> None:
        self.alias = alias
        self.name = name
        self.arguments = arguments
        self.directives = directives
        self.selections = selections

    def visit(self, visitor: QueryVisitor) -> None:
        visitor.visit_field_selection_begin(self)

        for arg in self.arguments:
            arg.visit(visitor)

        for directive in self.directives:
            directive.visit(visitor)

        for sel in self.selections:
            sel.visit(visitor)

        visitor.visit_field_selection_end(self)

    def to_primitive(self) -> PrimitiveType:
        d: Dict[str, PrimitiveType] = {"@f": self.name}

        add_if_not_none(d, "alias", self.alias)
        add_if_not_empty(d, "arguments", self.arguments)
        add_if_not_empty(d, "directives", self.directives)
        add_if_not_empty(d, "selections", self.selections)
        return d


class FragmentSpread(Selection):
    fragment_name: str
    directives: Sequence[Directive]

    def __init__(self, fragment_name: str, directives: Sequence[Directive]) -> None:
        self.fragment_name = fragment_name
        self.directives = directives

    def visit(self, visitor: QueryVisitor) -> None:
        visitor.visit_fragment_spread_begin(self)

        for directive in self.directives:
            directive.visit(visitor)

        visitor.visit_fragment_spread_end(self)

    def to_primitive(self) -> PrimitiveType:
        d: Dict[str, PrimitiveType] = {"@frg-spread": self.fragment_name}
        add_if_not_empty(d, "directives", self.directives)
        return d


class InlineFragment(Selection, Fragment):
    on_type: Optional[NamedType]
    directives: Sequence[Directive]
    selections: Sequence[Selection]

    def __init__(self,
                 on_type: Optional[NamedType],
                 directives: Sequence[Directive],
                 selections: Sequence[Selection]) -> None:
        self.on_type = on_type
        self.directives = directives
        self.selections = selections

    def visit(self, visitor: QueryVisitor) -> None:
        visitor.visit_inline_fragment_begin(self)

        for directive in self.directives:
            directive.visit(visitor)

        for sel in self.selections:
            sel.visit(visitor)

        visitor.visit_inline_fragment_end(self)

    def to_primitive(self) -> PrimitiveType:
        d: Dict[str, PrimitiveType] = {"@frg-inline": None}

        if self.on_type is not None:
            d["on_type"] = self.on_type.to_primitive()

        add_if_not_empty(d, "directives", self.directives)
        add_if_not_empty(d, "selections", self.selections)
        return d


__all__ = ["GraphQlModelType", "Document", "Operation", "Query", "Mutation", "VariableDefinition",
           "Type", "NamedType", "ListType", "Directive", "ConstValue", "Value", "Variable",
           "NullValue", "EnumValue", "IntValue", "FloatValue", "StrValue", "BoolValue",
           "ConstListValue", "ConstObjectValue", "ObjectValue", "Argument", "Fragment", "NamedFragment",
           "Selection", "FieldSelection", "FragmentSpread", "InlineFragment", "QueryVisitor"]



import json
import typing as t
from itertools import chain

import gql_alchemy.types as gt
from .errors import GqlSchemaError
from .utils import PrimitiveType

Boolean = gt.Boolean
Int = gt.Int
Float = gt.Float
String = gt.String
ID = gt.ID


def NonNull(of_type: t.Union[gt.ScalarType, gt.List, str]) -> gt.NonNull:
    return gt.NonNull(of_type)


def List(of_type: t.Union[gt.InlineType, str]) -> gt.List:
    return gt.List(of_type)


class TypeDefinition:
    def __init__(self, name: str, description: t.Optional[str] = None) -> None:
        self.name = name
        self.description = description

    def to_type(self) -> gt.UserType:
        raise NotImplementedError()

    def to_pretty_str(self) -> str:
        raise NotImplementedError()


class EnumValue:
    def __init__(self, name: str, description: t.Optional[str] = None,
                 is_deprecated: bool = False, deprecation_reason: t.Optional[str] = None) -> None:
        self.name = name
        self.description = description
        self.is_deprecated = is_deprecated
        self.deprecation_reason = deprecation_reason


class Enum(TypeDefinition):
    def __init__(self, name: str, enum_values: t.Sequence[t.Union[EnumValue, str]],
                 description: t.Optional[str] = None) -> None:
        TypeDefinition.__init__(self, name, description)
        self.values: t.Sequence[EnumValue] = [EnumValue(i) if isinstance(i, str) else i for i in enum_values]
        self.__names = {v.name for v in self.values}

    def to_type(self) -> gt.UserType:
        return gt.Enum(self.name, {v.name for v in self.values})

    def to_pretty_str(self) -> str:
        if self.name is None:
            raise RuntimeError("Name is not filled in")

        lines: t.List[str] = []

        _format_description(lines, 0, self.description)

        lines.append("enum {} {{".format(self.name))

        for v in self.values:
            _format_description(lines, 2, v.description, v.is_deprecated, v.deprecation_reason)
            lines.append("  " + v.name)

        lines.append("}")

        return "\n".join(lines)


class InputValue:
    def __init__(self, type: t.Union[gt.InlineType, str],
                 default_value: PrimitiveType = None,
                 description: t.Optional[str] = None) -> None:
        self.type = type
        self.default_value = default_value
        self.description = description

    def to_type_arg(self) -> gt.Argument:
        return gt.Argument(self.type, self.default_value)


class Field:
    def __init__(self, type: t.Union[gt.InlineType, str],
                 args: t.Mapping[str, t.Union[InputValue, gt.InlineType, str]] = {},
                 description: t.Optional[str] = None,
                 is_deprecated: bool = False,
                 deprecation_reason: t.Optional[str] = None) -> None:
        self.type = type

        norm_args: t.Dict[str, InputValue] = {}
        for name, arg in args.items():
            if isinstance(arg, InputValue):
                norm_args[name] = arg
            else:
                norm_args[name] = InputValue(arg)

        self.args: t.Mapping[str, InputValue] = norm_args
        self.description = description
        self.is_deprecated = is_deprecated
        self.deprecation_reason = deprecation_reason

    def to_type_field(self) -> gt.Field:
        return gt.Field(self.type, dict(((arg_name, arg.to_type_arg()) for arg_name, arg in self.args.items())))

    def format(self, name: str) -> str:
        lines: t.List[str] = []

        if self.description is not None or self.is_deprecated:
            _format_description(lines, 2, self.description, self.is_deprecated, self.deprecation_reason)

        if self.args is not None:
            if self.description is not None or self.is_deprecated:
                lines.append("  #")
            _format_args_descriptions(lines, self.args)

        lines.append("  " + name + _format_args(self.args) + ": " + str(self.type))

        return "\n".join(lines)


class _SelectableTypeDefinition(TypeDefinition):
    """Object, Interface"""

    def __init__(self, name: str,
                 fields: t.Mapping[str, t.Union[Field, gt.InlineType, str]],
                 description: t.Optional[str]) -> None:
        super().__init__(name, description)

        norm_fields: t.Dict[str, Field] = {}

        for name, field in fields.items():
            if isinstance(field, Field):
                norm_fields[name] = field
            else:
                norm_fields[name] = Field(field)

        self.fields: t.Mapping[str, Field] = norm_fields

    def to_type(self) -> gt.UserType:
        raise NotImplementedError()

    def to_pretty_str(self) -> str:
        lines: t.List[str] = []

        if self.description is not None:
            _format_description(lines, 0, self.description)

        lines.append(self._format_first_line())

        for n, f in self.fields.items():
            lines.append(f.format(n))

        lines.append("}")
        lines.append("")

        return "\n".join(lines)

    def _format_first_line(self) -> str:
        raise NotImplementedError()


class Object(_SelectableTypeDefinition):
    def __init__(self, name: str, fields: t.Mapping[str, t.Union[Field, gt.InlineType, str]],
                 interfaces: t.Set[str] = set(),
                 description: t.Optional[str] = None) -> None:
        super().__init__(name, fields, description)

        self.interfaces = interfaces

    def to_type(self) -> gt.Object:
        return gt.Object(self.name,
                         dict(((field_name, field.to_type_field()) for field_name, field in self.fields.items())),
                         self.interfaces)

    def _format_first_line(self) -> str:
        first_line = "type " + self.name

        if len(self.interfaces) > 0:
            first_line += " implements " + ", ".join(self.interfaces) + " {"

        return first_line


class Interface(_SelectableTypeDefinition):
    def __init__(self, name: str, fields: t.Mapping[str, t.Union[Field, gt.InlineType, str]],
                 description: t.Optional[str] = None) -> None:
        super().__init__(name, fields, description)

    def to_type(self) -> gt.UserType:
        return gt.Interface(self.name,
                            dict(((field_name, field.to_type_field()) for field_name, field in self.fields.items())))

    def _format_first_line(self) -> str:
        return "interface " + self.name + " {"


class Union(TypeDefinition):
    def __init__(self, name: str, possible_types: t.Set[str], description: t.Optional[str] = None) -> None:
        super().__init__(name, description)
        self.possible_types = possible_types

    def to_type(self) -> gt.UserType:
        return gt.Union(self.name, self.possible_types)

    def to_pretty_str(self) -> str:
        return "union " + self.name + " = " + " | ".join(self.possible_types)


class IoField:
    def __init__(self, type: t.Union[gt.InlineType, str], description: t.Optional[str] = None) -> None:
        self.type = type
        self.description = description


class InputObject(TypeDefinition):
    def __init__(self, name: str, input_fields: t.Mapping[str, t.Union[IoField, gt.InlineType, str]],
                 description: t.Optional[str] = None) -> None:
        super().__init__(name, description)

        norm_infs: t.Dict[str, IoField] = {}

        if len(input_fields) == 0:
            raise GqlSchemaError("Input object must have at least one field")

        for name, inf in input_fields.items():
            if isinstance(inf, IoField):
                norm_infs[name] = inf
            else:
                norm_infs[name] = IoField(inf)

        self.input_fields: t.Mapping[str, IoField] = norm_infs

    def to_type(self) -> gt.InputObject:
        return gt.InputObject(
            self.name,
            dict(((field_name, field.type) for field_name, field in self.input_fields.items()))
        )

    def to_pretty_str(self) -> str:
        lines: t.List[str] = []

        if self.description is not None:
            _format_description(lines, 0, self.description)

        lines.append("input " + self.name + "{")

        for n, f in self.input_fields.items():
            if f.description is not None:
                _format_description(lines, 2, f.description)

            f_str = "  " + n + ": " + str(f.type)

            lines.append(f_str)

        lines.append("}")
        lines.append("")

        return "\n".join(lines)


DirectiveLocations = gt.DirectiveLocations


class Directive:
    def __init__(self, name: str, locations: t.Set[gt.DirectiveLocation],
                 args: t.Optional[t.Mapping[str, t.Union[InputValue, gt.InlineType, str]]] = None,
                 description: t.Optional[str] = None) -> None:
        self.name = name
        self.locations = locations

        norm_args: t.Dict[str, InputValue] = {}
        if args is not None:
            for name, arg in args:
                if isinstance(arg, InputValue):
                    norm_args[name] = arg
                else:
                    norm_args[name] = InputValue(arg)

        self.args: t.Mapping[str, InputValue] = norm_args

        self.description = description

    def to_directive(self) -> gt.Directive:
        return gt.Directive(self.name, self.locations,
                            dict(((arg_name, arg.to_type_arg()) for arg_name, arg in self.args.items())))

    def to_pretty_str(self) -> str:
        lines: t.List[str] = []

        if self.description is not None:
            _format_description(lines, 0, self.description)

        if self.args is not None:
            if self.description is not None:
                lines.append("#")
            _format_args_descriptions(lines, self.args)

        lines.append("directive " + self.name + _format_args(self.args))
        lines.append("")

        return "\n".join(lines)


class Schema:
    def __init__(self,
                 types: t.Sequence[t.Union[Enum, Object, Interface, Union, InputObject]],
                 query: Object,
                 mutation: t.Optional[Object] = None,
                 directives: t.Optional[t.Sequence[Directive]] = None) -> None:

        self.types = list(types)
        self.query_object_name = query.name

        self.mutation_object_name: t.Optional[str] = None
        if mutation is not None:
            self.types.append(mutation)
            self.mutation_object_name = mutation.name

        self.directives: t.Sequence[Directive] = directives or []
        self.type_registry = gt.TypeRegistry(
            list(chain(
                (type_def.to_type() for type_def in self.types),
                [_IntrospectionQueryObject(query.to_type())]
            )),
            [dir_def.to_directive() for dir_def in self.directives]
        )

        self.types.append(query)

    def format(self) -> str:
        lines: t.List[str] = []

        for type_def in self.types:
            lines.append(type_def.to_pretty_str())

        if self.directives is not None:
            for d in self.directives:
                lines.append(d.to_pretty_str())

        return "\n".join(lines)


def _format_description(lines: t.List[str],
                        indent: int,
                        description: t.Optional[str],
                        is_deprecated: bool = False,
                        deprecation_reason: t.Optional[str] = None) -> None:
    if description is not None:
        for l in description.splitlines(keepends=False):
            lines.append(" " * indent + "# " + l)

    if is_deprecated:
        if description is not None:
            lines.append(" " * indent + "#")
        lines.append(" " * indent + "# Deprecated!")

    if deprecation_reason is not None:
        for l in deprecation_reason.splitlines(keepends=False):
            lines.append(" " * indent + "# " + l)


def _format_args_descriptions(lines: t.List[str], args: t.Mapping[str, InputValue]) -> None:
    for n, a in args.items():
        if a.description is None:
            lines.append("  # " + n + ": undocumented")
        else:
            lines.append("  # " + n + ":")
            for l in a.description.splitlines(keepends=False):
                lines.append("  #   " + l)


def _format_args(args: t.Optional[t.Mapping[str, InputValue]]) -> str:
    if args is None:
        return ""

    parts = ["("]

    first = True

    for n, a in args.items():

        if not first:
            parts.append(", ")
        first = False

        parts.append(n)
        parts.append(": ")
        parts.append(str(a.type))
        if a.default_value is not None:
            parts.append(" = ")
            parts.append(json.dumps(a.default_value))

    parts.append(")")

    return "".join(parts)


class _IntrospectionQueryObject(gt.Object):
    def __init__(self, query_obj: gt.Object) -> None:
        super().__init__(str(query_obj), {
            "__schema": gt.Field(gt.NonNull("__Schema"), {}),
            "__type": gt.Field("__Type", {"name": gt.Argument(gt.NonNull(gt.String), None)})
        }, set())

        self.__query_obj = query_obj

    def fields(self, type_registry: gt.TypeRegistry) -> t.Mapping[str, gt.Field]:
        return dict(
            chain(
                self.__query_obj.fields(type_registry).items(),
                super().fields(type_registry).items()
            )
        )

    def own_fields(self) -> t.Mapping[str, gt.Field]:
        return dict(
            chain(
                self.__query_obj.own_fields().items(),
                super().own_fields().items()
            )
        )

    def implements(self, type_registry: gt.TypeRegistry) -> t.Sequence[gt.Interface]:
        return self.__query_obj.implements(type_registry)


__all__ = ["Boolean", "Int", "Float", "String", "ID", "NonNull", "List", "EnumValue", "Enum", "InputValue",
           "Field", "Object", "Interface", "Union", "IoField", "InputObject", "DirectiveLocations", "Directive",
           "Schema"]



import json
import logging
import re
import typing as t

import gql_alchemy.query_model as qm
from .errors import GqlParsingError
from .raw_reader import Reader, format_position
from .utils import PrimitiveType, add_if_not_empty, add_if_not_none

logger = logging.getLogger("gql_alchemy")


def log_stack(stack: t.Sequence['ElementParser']) -> None:
    if not logger.isEnabledFor(logging.DEBUG):
        return

    logger.debug("Stack: %s", json.dumps([i.to_dbg_repr() for i in stack]))


def log_position(reader: Reader) -> None:
    if not logger.isEnabledFor(logging.DEBUG):
        return

    lineno = reader.lineno
    line_pos = reader.line_pos()
    lines = [reader.prev_line(), reader.current_line(), reader.next_line()]

    for line in format_position(lineno, line_pos, lines):
        logger.debug(line)


class VerifyDocument(qm.QueryVisitor):
    def __init__(self) -> None:
        self.__current_op_name: t.Optional[str] = None
        self.__current_fragment_name: t.Optional[str] = None
        self.__declared_vars: t.Dict[str, t.MutableSet[str]] = {}
        self.__o2f_calls: t.Dict[str, t.MutableSet[str]] = {}
        self.__f2f_calls: t.Dict[str, t.MutableSet[str]] = {}
        self.__direct_fragments_variables: t.Dict[str, t.MutableSet[str]] = {}
        self.__fragments_variables: t.Dict[str, t.Dict[str, t.MutableSet[str]]] = {}  # fragment -> (var -> fragment)

    def visit_query_begin(self, query: qm.Query) -> None:
        self.__current_op_name = query.name if query.name is not None else "!non-named"
        self.__o2f_calls[self.__current_op_name] = set()
        self.__declared_vars[self.__current_op_name] = {v.name for v in query.variables}

    def visit_query_end(self, query: qm.Query) -> None:
        self.__current_op_name = None

    def visit_mutation_begin(self, mutation: qm.Mutation) -> None:
        self.__current_op_name = mutation.name if mutation.name is not None else "!non-named"
        self.__o2f_calls[self.__current_op_name] = set()
        self.__declared_vars[self.__current_op_name] = {v.name for v in mutation.variables}

    def visit_mutation_end(self, mutation: qm.Mutation) -> None:
        self.__current_op_name = None

    def visit_fragment_begin(self, fragment: qm.NamedFragment) -> None:
        self.__direct_fragments_variables[fragment.name] = set()
        self.__f2f_calls[fragment.name] = set()
        self.__current_fragment_name = fragment.name

    def visit_fragment_end(self, fragment: qm.NamedFragment) -> None:
        self.__current_fragment_name = None

    def visit_variable(self, var: qm.Variable) -> None:
        if self.__current_op_name is not None:
            if var.name not in self.__declared_vars[self.__current_op_name]:
                raise GqlParsingError("Undefined variable `{}` used in `{}` operation".format(
                    var.name, self.__current_op_name
                ))
            return

        if self.__current_fragment_name is None:
            raise RuntimeError("Operation or fragment name expected here")

        self.__direct_fragments_variables[self.__current_fragment_name].add(var.name)

    def visit_fragment_spread_begin(self, spread: qm.FragmentSpread) -> None:
        if self.__current_fragment_name is None:
            if self.__current_op_name is None:
                raise RuntimeError("Fragment or operation name expected here")
            self.__o2f_calls[self.__current_op_name].add(spread.fragment_name)
            return

        self.__f2f_calls[self.__current_fragment_name].add(spread.fragment_name)

    def visit_document_end(self, document: qm.Document) -> None:
        for op, frs in self.__o2f_calls.items():
            for fr in frs:
                if fr not in self.__f2f_calls:
                    raise GqlParsingError("Undefined fragment `{}` used in `{}` operation".format(fr, op))

                if fr not in self.__fragments_variables:
                    self.__check_cycles_and_collect_variables(fr, set())

                self.__verify_fragment_vars(op, fr)

        called_from_ops = set(self.__fragments_variables.keys())
        defined = set(self.__direct_fragments_variables.keys())
        unused = list(defined.difference(called_from_ops))
        unused.sort()
        if len(unused) > 0:
            raise GqlParsingError("Unused fragments detected: {}".format(
                ", ".join(('`' + f + '`' for f in unused))
            ))

    def __check_cycles_and_collect_variables(self, fr: str, visited: t.MutableSet[str]) -> None:
        if fr in self.__fragments_variables:
            return

        visited.add(fr)

        self.__fragments_variables[fr] = {}
        for var in self.__direct_fragments_variables[fr]:
            self.__fragments_variables[fr][var] = set()
            self.__fragments_variables[fr][var].add(fr)

        for called_fr in self.__f2f_calls[fr]:
            if called_fr in visited:
                raise GqlParsingError("Cycled fragment usage detected for `{}` fragment".format(fr))
            self.__check_cycles_and_collect_variables(called_fr, visited)
            for var, used_in in self.__fragments_variables[called_fr].items():
                if var not in self.__fragments_variables[fr]:
                    self.__fragments_variables[fr][var] = set()
                for user_in_fr in used_in:
                    self.__fragments_variables[fr][var].add(user_in_fr)

        visited.remove(fr)

    def __verify_fragment_vars(self, op_name: str, fragment_name: str) -> None:
        fragment_variables = self.__fragments_variables[fragment_name]
        for var, used_in in fragment_variables.items():
            if var not in self.__declared_vars[op_name]:
                if len(used_in) == 1:
                    used_in_str = "`{}` fragment".format(next(iter(used_in)))
                else:
                    used_in_list = list(used_in)
                    used_in_list.sort()
                    used_in_str = "{} fragments".format(", ".join(('`' + i + '`' for i in used_in_list)))
                raise GqlParsingError(
                    "Undefined variable `{}` used in {} when called from `{}` operation".format(
                        var, used_in_str, op_name
                    )
                )


def parse_document(text_input: str) -> qm.Document:
    document: t.List[qm.Document] = []

    def set_document(d: qm.Document) -> None:
        document.append(d)

    parser = DocumentParser(set_document)

    parse(text_input, parser)

    return document[0]


def push_to_stack(stack: t.List['ElementParser'], parser: 'ElementParser', reader: Reader) -> None:
    stack.append(parser)

    log_stack(stack)

    index_before = reader.index
    parser.consume(reader)
    index_after = reader.index

    if index_after != index_before:
        logger.debug("READ DETECTED")
        log_position(reader)
        log_stack(stack)


def parse(text_input: str, initial_parser: 'ElementParser') -> None:
    stack: t.List[ElementParser] = []
    reader = Reader(text_input)

    logger.debug("=== START PARSING ===")
    logger.debug("INPUT\n%s\n===", text_input)
    push_to_stack(stack, initial_parser, reader)

    while True:
        if len(stack) == 0:
            logger.debug("=== FINISH PARSING ===")
            log_stack(stack)

            if reader.lookup_ch() is not None:
                log_position(reader)
                raise GqlParsingError("Not all input parsed", reader)

            logger.debug("EOI")
            return

        parser = stack[-1]

        index_before = reader.index
        to_add, remove_count = parser.next(reader)
        index_after = reader.index

        if to_add is None and remove_count == 0 and index_before == index_after:
            parser_name = type(parser).__name__
            logger.debug("Parser did nothing: %s", parser_name)
            raise RuntimeError("Parser did no change during a step, stop "
                               "parsing to prevent looping forever (parser: {})".format(parser_name))

        if index_after != index_before:
            logger.debug("READ DETECTED")
            log_position(reader)
            log_stack(stack)

        if remove_count > 0:
            stack[-remove_count:] = []
            log_stack(stack)

        if to_add is not None:
            push_to_stack(stack, to_add, reader)


class LiteralExpected(GqlParsingError):
    def __init__(self, symbols: t.Sequence[str], reader: Reader) -> None:
        if len(symbols) == 1:
            msg = "Expected '{}'".format(symbols[0])
        else:
            msg = "One of {} and {} expected".format(
                ', '.join(('"{}"'.format(s) for s in symbols[:-1])),
                '"{}"'.format(symbols[-1]))
        super().__init__(msg, reader)
        self.symbols = symbols


NAME_RE = re.compile(r'[_A-Za-z][_0-9A-Za-z]*')


class ElementParser:
    @staticmethod
    def assert_ch(reader: Reader, ch: str) -> None:
        next_ch = reader.read_ch()
        if next_ch != ch:
            raise LiteralExpected([ch], reader)

    @staticmethod
    def assert_literal(reader: Reader, literal: str) -> None:
        next_literal = reader.read_re(re.compile(r'(?:' + literal + r')(?=[^_0-9A-Za-z]|$)'))
        if next_literal is None:
            raise LiteralExpected([literal], reader)

    @staticmethod
    def assert_ellipsis(reader: Reader) -> None:
        next_literal = reader.read_re(re.compile(r'[.]{3}(?=[^.]|$)'))
        if next_literal is None:
            raise LiteralExpected(["..."], reader)

    @staticmethod
    def try_literal(reader: Reader, literal: str) -> bool:
        next_literal = reader.read_re(re.compile(r'(?:' + literal + r')(?=[^_0-9A-Za-z]|$)'))
        return next_literal is not None

    @staticmethod
    def read_if(reader: Reader, ch: str) -> bool:
        next_ch = reader.lookup_ch()

        if next_ch == ch:
            reader.read_ch()
            return True

        return False

    @staticmethod
    def read_name(reader: Reader) -> str:
        name = reader.read_re(NAME_RE)

        if name is None:
            raise GqlParsingError("Name expected", reader)

        return name

    def consume(self, reader: Reader) -> None:
        raise NotImplementedError()

    def next(self, reader: Reader) -> t.Tuple[t.Optional['ElementParser'], int]:
        raise NotImplementedError()

    def to_dbg_repr(self) -> PrimitiveType:
        return {
            "parser": type(self).__name__
        }


class DocumentParser(ElementParser):
    def __init__(self, set_document: t.Callable[[qm.Document], None]) -> None:
        self.set_document = set_document

        self.operations: t.List[qm.Operation] = []
        self.fragments: t.List[qm.NamedFragment] = []
        self.selections: t.List[qm.Selection] = []

        self.selection_allowed = True
        self.query_allowed = True

    def consume(self, reader: Reader) -> None:
        pass

    def next(self, reader: Reader) -> t.Tuple[t.Optional[ElementParser], int]:
        if len(self.selections) > 0:
            self.operations.append(qm.Query(None, [], [], self.selections))
            self.selections = []

        ch = reader.lookup_ch()

        if self.selection_allowed and ch == "{":
            self.selection_allowed = False
            self.query_allowed = False
            return SelectionsParser(self.selections, None), 0

        if ch == "m":
            return OperationParser("mutation", self.operations), 0

        if self.query_allowed and ch == "q":
            self.selection_allowed = False
            return OperationParser("query", self.operations), 0

        if ch == "f":
            return FragmentParser(self.fragments), 0

        if ch is None:
            self.__verify_and_set_document(qm.Document(self.operations, self.fragments))
            return None, 1

        raise GqlParsingError("One of top-level declaration expected", reader)

    def __verify_and_set_document(self, document: qm.Document) -> None:
        document.visit(VerifyDocument())

        self.set_document(document)

    def to_dbg_repr(self) -> PrimitiveType:
        d = t.cast(t.Dict[str, PrimitiveType], super().to_dbg_repr())
        add_if_not_empty(d, "ops", self.operations)
        add_if_not_empty(d, "frgs", self.fragments)
        add_if_not_empty(d, "sels", self.selections)
        return d


class OperationParser(ElementParser):
    variables: t.List[qm.VariableDefinition]
    directives: t.List[qm.Directive]
    selections: t.List[qm.Selection]

    def __init__(self, operation_type: str, operations: t.List[qm.Operation]) -> None:
        self.operation_type = operation_type
        self.operations = operations

        self.name: t.Optional[str] = None
        self.variables = []
        self.directives = []
        self.selections = []

        self.expected = {
            '(': self.next_variables,
            '@': self.next_directives,
            '{': self.next_selections
        }

    def consume(self, reader: Reader) -> None:
        self.assert_literal(reader, self.operation_type)

        ch = reader.lookup_ch()

        if ch not in self.expected:
            name = reader.read_re(NAME_RE)
            if name is None:
                raise GqlParsingError("One of `name`, '(', '@' or '{' expected", reader)
            self.name = name

        if self.name in {op.name for op in self.operations}:
            raise GqlParsingError("Operation with the same name already exists", reader)

    def next(self, reader: Reader) -> t.Tuple[t.Optional[ElementParser], int]:

        if len(self.expected) == 0:
            if self.operation_type == "query":
                self.operations.append(qm.Query(self.name, self.variables, self.directives, self.selections))
            else:
                self.operations.append(qm.Mutation(self.name, self.variables, self.directives, self.selections))
            return None, 1

        ch = reader.lookup_ch()

        if ch not in self.expected:
            raise LiteralExpected(list(self.expected.keys()), reader)

        return self.expected[ch]()

    def next_variables(self) -> t.Tuple[ElementParser, int]:
        del self.expected['(']
        return VariablesParser(self.variables), 0

    def next_directives(self) -> t.Tuple[ElementParser, int]:
        if '(' in self.expected:
            del self.expected['(']
        del self.expected['@']
        return DirectivesParser(self.directives), 0

    def next_selections(self) -> t.Tuple[ElementParser, int]:
        self.expected.clear()
        return SelectionsParser(self.selections, None), 0

    def to_dbg_repr(self) -> PrimitiveType:
        d: t.Dict[str, PrimitiveType] = {
            "type": self.operation_type
        }
        add_if_not_none(d, "name", self.name)
        add_if_not_empty(d, "vars", self.variables)
        add_if_not_empty(d, "dirs", self.directives)
        add_if_not_empty(d, "sels", self.selections)
        return d


class VariablesParser(ElementParser):
    def __init__(self, variables: t.List[qm.VariableDefinition]) -> None:
        self.variables = variables

    def consume(self, reader: Reader) -> None:
        self.assert_ch(reader, "(")

    def next(self, reader: Reader) -> t.Tuple[t.Optional[ElementParser], int]:
        ch = reader.lookup_ch()

        if ch == "$":
            return VariableDefinitionParser(self.variables), 0

        if ch == ")":
            if len(self.variables) == 0:
                raise GqlParsingError("Empty variable definition is not allowed", reader)
            reader.read_ch()
            return None, 1

        raise LiteralExpected(["$", ")"], reader)


class VariableDefinitionParser(ElementParser):
    type: t.Optional[qm.Type]

    def __init__(self, variables: t.List[qm.VariableDefinition]) -> None:
        self.variables = variables

        self.name: t.Optional[str] = None
        self.type: t.Optional[qm.Type] = None
        self.default: t.Optional[qm.ConstValue] = None

        self.default_checked = False

    def consume(self, reader: Reader) -> None:
        self.assert_ch(reader, "$")

        self.name = self.read_name(reader)

        if self.name in {v.name for v in self.variables}:
            raise GqlParsingError("Variable with the same name already defined", reader)

        self.assert_ch(reader, ":")

        self.type = self.parse_type(reader)

    def parse_type(self, reader: Reader) -> qm.Type:
        ch = reader.lookup_ch()

        if ch == "[":
            return self.parse_list_type(reader)

        type_name = self.read_name(reader)

        return qm.NamedType(type_name, not self.read_if(reader, "!"))

    def parse_list_type(self, reader: Reader) -> qm.ListType:
        self.assert_ch(reader, "[")

        el_type = self.parse_type(reader)

        self.assert_ch(reader, "]")

        return qm.ListType(el_type, not self.read_if(reader, "!"))

    def next(self, reader: Reader) -> t.Tuple[t.Optional[ElementParser], int]:
        if not self.default_checked:
            self.default_checked = True

            if self.read_if(reader, "="):
                def set_default(v: qm.ConstValue) -> None:
                    self.default = v

                return ConstValueParser(set_default), 0

        if self.name is None or self.type is None:
            raise RuntimeError("Unexpected `None`")

        self.variables.append(qm.VariableDefinition(self.name, self.type, self.default))

        return None, 1

    def to_dbg_repr(self) -> PrimitiveType:
        d = t.cast(t.Dict[str, PrimitiveType], super().to_dbg_repr())
        add_if_not_none(d, "name", self.name)

        if self.type is not None:
            d["type"] = self.type.to_primitive()

        if self.default is not None:
            d["default"] = self.default.to_primitive()
        return d


class DirectivesParser(ElementParser):
    def __init__(self, directives: t.List[qm.Directive]) -> None:
        self.directives = directives

    def consume(self, reader: Reader) -> None:
        pass

    def next(self, reader: Reader) -> t.Tuple[t.Optional[ElementParser], int]:
        if reader.lookup_ch() == "@":
            return DirectiveParser(self.directives), 0

        return None, 1


class DirectiveParser(ElementParser):
    def __init__(self, directives: t.List[qm.Directive]) -> None:
        self.directives = directives

        self.name: t.Optional[str] = None
        self.arguments: t.List[qm.Argument] = []

        self.arguments_checked = False

    def consume(self, reader: Reader) -> None:
        self.assert_ch(reader, "@")

        self.name = self.read_name(reader)

    def next(self, reader: Reader) -> t.Tuple[t.Optional[ElementParser], int]:
        if not self.arguments_checked:
            self.arguments_checked = True

            if reader.lookup_ch() == "(":
                return ArgumentsParser(self.arguments), 0

        if self.name is None:
            raise RuntimeError("Unexpected `None`")

        self.directives.append(qm.Directive(self.name, self.arguments))
        return None, 1


class ArgumentsParser(ElementParser):
    def __init__(self, arguments: t.List[qm.Argument]) -> None:
        self.arguments = arguments

    def consume(self, reader: Reader) -> None:
        self.assert_ch(reader, "(")

    def next(self, reader: Reader) -> t.Tuple[t.Optional[ElementParser], int]:
        if self.read_if(reader, ")"):
            if len(self.arguments) == 0:
                raise GqlParsingError("Empty arguments list is not allowed", reader)

            return None, 1

        return ArgumentParser(self.arguments), 0


class ArgumentParser(ElementParser):
    def __init__(self, arguments: t.List[qm.Argument]) -> None:
        self.arguments = arguments

        self.name: t.Optional[str] = None
        self.value: t.Optional[qm.Value] = None

        self.value_parsed = False

    def consume(self, reader: Reader) -> None:
        self.name = self.read_name(reader)

        if self.name in {a.name for a in self.arguments}:
            raise GqlParsingError("Argument with the same name already defined", reader)

        self.assert_ch(reader, ":")

    def next(self, reader: Reader) -> t.Tuple[t.Optional[ElementParser], int]:
        if self.value_parsed:

            if self.name is None or self.value is None:
                raise RuntimeError("Unexpected `None`")

            self.arguments.append(qm.Argument(self.name, self.value))

            return None, 1

        self.value_parsed = True

        def set_value(v: qm.Value) -> None:
            self.value = v

        return ValueParser(set_value), 0

    def to_dbg_repr(self) -> PrimitiveType:
        d = t.cast(t.Dict[str, PrimitiveType], super().to_dbg_repr())

        add_if_not_none(d, "name", self.name)

        if self.value is not None:
            d["value"] = self.value.to_primitive()

        return d


class SelectionsParser(ElementParser):
    DETECT_FRAGMENT_SPREAD_RE = re.compile(r'[.]{3}[ \t]*([_A-Za-z][_0-9A-Za-z]*)')

    def __init__(self, selections: t.List[qm.Selection], selected_aliases: t.Optional[t.MutableSet[str]]) -> None:
        self.selections = selections
        self.selected_aliases: t.MutableSet[str] = selected_aliases if selected_aliases is not None else set()

    def consume(self, reader: Reader) -> None:
        self.assert_ch(reader, "{")

    def next(self, reader: Reader) -> t.Tuple[t.Optional[ElementParser], int]:
        if self.read_if(reader, "}"):

            if len(self.selections) == 0:
                raise GqlParsingError("Empty selection set is not allowed", reader)

            return None, 1

        if reader.lookup_ch() == ".":
            m = reader.match_re(self.DETECT_FRAGMENT_SPREAD_RE)
            if m is not None and m.group(1) != "on":
                return FragmentSpreadParser(self.selections), 0
            return InlineFragmentParser(self.selections, self.selected_aliases), 0

        return FieldParser(self.selections, self.selected_aliases), 0


class FieldParser(ElementParser):
    def __init__(self, selections: t.List[qm.Selection], selected_aliases: t.MutableSet[str]) -> None:
        self.parent_selections = selections
        self.selected_aliases = selected_aliases

        self.alias: t.Optional[str] = None
        self.name: t.Optional[str] = None
        self.arguments: t.List[qm.Argument] = []
        self.directives: t.List[qm.Directive] = []
        self.selections: t.List[qm.Selection] = []

        self.can_be = {
            '(': self.next_arguments,
            '@': self.next_directives,
            '{': self.next_selections
        }

    def consume(self, reader: Reader) -> None:
        name = self.read_name(reader)

        if name in self.selected_aliases:
            raise GqlParsingError("Selection under `{}` alias already defined".format(name), reader)
        self.selected_aliases.add(name)

        if self.read_if(reader, ":"):
            self.alias = name
            self.name = self.read_name(reader)
        else:
            self.name = name

    def next(self, reader: Reader) -> t.Tuple[t.Optional[ElementParser], int]:
        ch = reader.lookup_ch()

        if ch in self.can_be:
            return self.can_be[ch]()

        if self.name is None:
            raise RuntimeError("Unexpected `None`")

        self.parent_selections.append(qm.FieldSelection(self.alias, self.name, self.arguments, self.directives,
                                                        self.selections))

        return None, 1

    def next_arguments(self) -> t.Tuple[ElementParser, int]:
        del self.can_be['(']
        return ArgumentsParser(self.arguments), 0

    def next_directives(self) -> t.Tuple[ElementParser, int]:
        if '(' in self.can_be:
            del self.can_be['(']
        del self.can_be['@']
        return DirectivesParser(self.directives), 0

    def next_selections(self) -> t.Tuple[ElementParser, int]:
        self.can_be.clear()
        return SelectionsParser(self.selections, None), 0

    def to_dbg_repr(self) -> PrimitiveType:
        d = t.cast(t.Dict[str, PrimitiveType], super().to_dbg_repr())
        add_if_not_none(d, "name", self.name)
        add_if_not_none(d, "alias", self.alias)
        add_if_not_empty(d, "args", self.arguments)
        add_if_not_empty(d, "dirs", self.directives)
        add_if_not_empty(d, "sels", self.selections)
        return d


ValueType = t.TypeVar("ValueType", qm.Value, qm.ConstValue)


class GenericValueParser(ElementParser, t.Generic[ValueType]):
    INT_PART = r'-?(?:[1-9][0-9]*|0)'
    FR_PART = r'(?:\.[0-9]+)'
    EXP_PART = r'(?:[eE][+-]?[0-9]+)'
    INT_RE = re.compile(INT_PART)
    FLOAT_RE = re.compile(INT_PART + r'(?:' + FR_PART + EXP_PART + '?|' + EXP_PART + ')')

    def __init__(self, set_value: t.Callable[[ValueType], None]) -> None:
        self.set_value: t.Callable[[ValueType], None] = set_value

    def consume(self, reader: Reader) -> None:
        pass

    def next(self, reader: Reader) -> t.Tuple[t.Optional[ElementParser], int]:
        ch = reader.lookup_ch()

        if ch == "$":
            self.parse_and_save_variable(reader)
            return None, 1

        if ch == '"':
            return StringValueParser(self.set_value), 1

        if ch == "[":
            return self.create_list_value_parser(), 1

        if ch == "{":
            return self.create_object_value_parser(), 1

        if self.try_literal(reader, "true"):
            self.set_value(qm.BoolValue(True))
            return None, 1

        if self.try_literal(reader, "false"):
            self.set_value(qm.BoolValue(False))
            return None, 1

        if self.try_literal(reader, "null"):
            self.set_value(qm.NullValue())
            return None, 1

        v = reader.read_re(NAME_RE)
        if v is not None:
            self.set_value(qm.EnumValue(v))
            return None, 1

        v = reader.read_re(self.FLOAT_RE)
        if v is not None:
            self.set_value(qm.FloatValue(float(v)))
            return None, 1

        v = reader.read_re(self.INT_RE)
        if v is not None:
            self.set_value(qm.IntValue(int(v)))
            return None, 1

        raise GqlParsingError("Value expected", reader)

    def parse_and_save_variable(self, reader: Reader) -> None:
        raise NotImplementedError()

    def create_list_value_parser(self) -> ElementParser:
        raise NotImplementedError()

    def create_object_value_parser(self) -> ElementParser:
        raise NotImplementedError()


class ValueParser(GenericValueParser[qm.Value]):
    def parse_and_save_variable(self, reader: Reader) -> None:
        self.assert_ch(reader, "$")
        name = self.read_name(reader)

        self.set_value(qm.Variable(name))

    def create_list_value_parser(self) -> ElementParser:
        return ListValueParser(self.set_value)

    def create_object_value_parser(self) -> ElementParser:
        return ObjectValueParser(self.set_value)


class ConstValueParser(GenericValueParser[qm.ConstValue]):
    def parse_and_save_variable(self, reader: Reader) -> None:
        # self.assert_ch(reader, "$")
        # name = self.read_name(reader)
        #
        # self.set_value(qm.Variable(name))
        raise GqlParsingError("Unexpected '$'", reader)

    def create_list_value_parser(self) -> ElementParser:
        return ConstListValueParser(self.set_value)

    def create_object_value_parser(self) -> ElementParser:
        return ConstObjectValueParser(self.set_value)


class GenericListValueParser(ElementParser, t.Generic[ValueType]):
    def __init__(self) -> None:
        self.values: t.List[ValueType] = []

    def consume(self, reader: Reader) -> None:
        self.assert_ch(reader, "[")

    def next(self, reader: Reader) -> t.Tuple[t.Optional[ElementParser], int]:
        if self.read_if(reader, "]"):
            self.save_values()

            return None, 1

        return self.create_value_parser(), 0

    def append_value(self, v: ValueType) -> None:
        self.values.append(v)

    def to_dbg_repr(self) -> PrimitiveType:
        d = t.cast(t.Dict[str, PrimitiveType], super().to_dbg_repr())
        add_if_not_empty(d, "values", self.values)
        return d

    def create_value_parser(self) -> ElementParser:
        raise NotImplementedError()

    def save_values(self) -> None:
        raise NotImplementedError()


class ListValueParser(GenericListValueParser[qm.Value]):
    def __init__(self, set_value: t.Callable[[qm.Value], None]) -> None:
        super().__init__()
        self.set_value = set_value

    def create_value_parser(self) -> ElementParser:
        return ValueParser(self.append_value)

    def save_values(self) -> None:
        self.set_value(qm.ListValue(self.values))


class ConstListValueParser(GenericListValueParser[qm.ConstValue]):
    def __init__(self, set_value: t.Callable[[qm.ConstValue], None]) -> None:
        super().__init__()
        self.set_value = set_value

    def create_value_parser(self) -> ElementParser:
        return ConstValueParser(self.append_value)

    def save_values(self) -> None:
        self.set_value(qm.ConstListValue(self.values))


class GenericObjectValueParser(ElementParser, t.Generic[ValueType]):
    def __init__(self) -> None:
        self.values: t.Dict[str, ValueType] = {}

    def consume(self, reader: Reader) -> None:
        self.assert_ch(reader, "{")

    def next(self, reader: Reader) -> t.Tuple[t.Optional[ElementParser], int]:
        if self.read_if(reader, "}"):
            self.save_values()

            return None, 1

        name = self.read_name(reader)

        self.assert_ch(reader, ":")

        return self.create_value_parser(name), 0

    def add_value(self, name: str) -> t.Callable[[ValueType], None]:
        def add(v: ValueType) -> None:
            self.values[name] = v

        return add

    def to_dbg_repr(self) -> PrimitiveType:
        d = t.cast(t.Dict[str, PrimitiveType], super().to_dbg_repr())
        d["values"] = dict(((k, v.to_primitive()) for k, v in self.values.items()))
        return d

    def save_values(self) -> None:
        raise NotImplementedError()

    def create_value_parser(self, name: str) -> ElementParser:
        raise NotImplementedError()


class ObjectValueParser(GenericObjectValueParser[qm.Value]):
    def __init__(self, set_value: t.Callable[[qm.Value], None]) -> None:
        super().__init__()
        self.set_value = set_value

    def save_values(self) -> None:
        self.set_value(qm.ObjectValue(self.values))

    def create_value_parser(self, name: str) -> ElementParser:
        return ValueParser(self.add_value(name))


class ConstObjectValueParser(GenericObjectValueParser[qm.ConstValue]):
    def __init__(self, set_value: t.Callable[[qm.ConstValue], None]) -> None:
        super().__init__()
        self.set_value = set_value

    def save_values(self) -> None:
        self.set_value(qm.ConstObjectValue(self.values))

    def create_value_parser(self, name: str) -> ElementParser:
        return ConstValueParser(self.add_value(name))


class StringValueParser(ElementParser):
    HEX_DIGITS = frozenset("0123456789abcdefABCDEF")

    def __init__(self, set_value: t.Callable[[qm.StrValue], None]) -> None:
        self.set_value = set_value

        self.value = ""

    def consume(self, reader: Reader) -> None:
        self.assert_ch(reader, '"')

    def next(self, reader: Reader) -> t.Tuple[t.Optional[ElementParser], int]:
        while True:
            ch = reader.str_read_ch()

            if ch == '"':
                self.set_value(qm.StrValue(self.value))
                return None, 1

            if ch == '\\':
                self.value += self.parse_escape(reader)
                continue

            if ch in {"\n", "\r"}:
                raise GqlParsingError("New line is not allowed in string literal", reader)

            if ch is None:
                raise GqlParsingError("Unexpected end of input", reader)

            self.value += ch

    def parse_escape(self, reader: Reader) -> str:
        ch = reader.str_read_ch()

        if ch == "u":
            return self.parse_unicode(reader)

        if ch == '"':
            return '"'

        if ch == '\\':
            return '\\'

        if ch == '/':
            return '/'

        if ch == 'b':
            return '\b'

        if ch == 'f':
            return '\f'

        if ch == 'n':
            return '\n'

        if ch == 'r':
            return '\r'

        if ch == 't':
            return '\t'

        raise GqlParsingError("Unexpected symbol '{}' after '\\' in string literal".format(ch), reader)

    def parse_unicode(self, reader: Reader) -> str:
        digits = []

        for i in range(4):
            digits.append(self.parse_hex_digit(reader))

        return chr(int(''.join(digits), 16))

    def parse_hex_digit(self, reader: Reader) -> str:
        ch = reader.str_read_ch()

        if ch in self.HEX_DIGITS:
            return ch

        raise GqlParsingError("Hex digit expected", reader)

    def to_dbg_repr(self) -> PrimitiveType:
        d = t.cast(t.Dict[str, PrimitiveType], super().to_dbg_repr())
        d["value"] = self.value
        return d


class FragmentSpreadParser(ElementParser):
    def __init__(self, selections: t.List[qm.Selection]) -> None:
        self.selections = selections

        self.name: t.Optional[str] = None
        self.directives: t.List[qm.Directive] = []

    def consume(self, reader: Reader) -> None:
        self.assert_ellipsis(reader)

        self.name = self.read_name(reader)

    def next(self, reader: Reader) -> t.Tuple[t.Optional[ElementParser], int]:
        if reader.lookup_ch() == '@':
            return DirectivesParser(self.directives), 0

        if self.name is None:
            raise RuntimeError("Unexpected `None`")

        self.selections.append(qm.FragmentSpread(self.name, self.directives))

        return None, 1

    def to_dbg_repr(self) -> PrimitiveType:
        d = t.cast(t.Dict[str, PrimitiveType], super().to_dbg_repr())
        add_if_not_none(d, "name", self.name)
        add_if_not_empty(d, "dirs", self.directives)
        return d


class InlineFragmentParser(ElementParser):
    on_type: t.Optional[qm.NamedType]

    def __init__(self, selections: t.List[qm.Selection], selected_aliases: t.MutableSet[str]) -> None:
        self.parent_selections = selections
        self.selected_aliases = selected_aliases

        self.on_type: t.Optional[qm.NamedType] = None
        self.directives: t.List[qm.Directive] = []
        self.selections: t.List[qm.Selection] = []

        self.directives_parsed = False
        self.selections_parsed = False

    def consume(self, reader: Reader) -> None:
        self.assert_ellipsis(reader)

        if self.try_literal(reader, "on"):
            type_name = self.read_name(reader)

            self.on_type = qm.NamedType(type_name, True)

    def next(self, reader: Reader) -> t.Tuple[t.Optional[ElementParser], int]:
        if not self.directives_parsed and reader.lookup_ch() == "@":
            self.directives_parsed = True
            return DirectivesParser(self.directives), 0

        if not self.selections_parsed:
            self.directives_parsed = True
            self.selections_parsed = True
            return SelectionsParser(self.selections, self.selected_aliases), 0

        self.parent_selections.append(qm.InlineFragment(self.on_type, self.directives, self.selections))
        return None, 1

    def to_dbg_repr(self) -> PrimitiveType:
        d = t.cast(t.Dict[str, PrimitiveType], super().to_dbg_repr())

        if self.on_type is not None:
            d["on_type"] = self.on_type.to_primitive()

        add_if_not_empty(d, "dirs", self.directives)
        add_if_not_empty(d, "sels", self.selections)

        return d


class FragmentParser(ElementParser):
    def __init__(self, fragments: t.List[qm.NamedFragment]) -> None:
        self.fragments = fragments

        self.name: t.Optional[str] = None
        self.on_type: t.Optional[qm.NamedType] = None
        self.directives: t.List[qm.Directive] = []
        self.selections: t.List[qm.Selection] = []

        self.directives_parsed = False
        self.selections_parsed = False

    def consume(self, reader: Reader) -> None:
        self.assert_literal(reader, "fragment")

        self.name = self.read_name(reader)

        if self.name == "on":
            raise GqlParsingError("Fragment can not have name \"on\"", reader)

        if self.name in {f.name for f in self.fragments}:
            raise GqlParsingError("Fragment with the same name already defined", reader)

        self.assert_literal(reader, "on")

        type_name = self.read_name(reader)

        self.on_type = qm.NamedType(type_name, True)

    def next(self, reader: Reader) -> t.Tuple[t.Optional[ElementParser], int]:
        if not self.directives_parsed and reader.lookup_ch() == "@":
            self.directives_parsed = True
            return DirectivesParser(self.directives), 0

        if not self.selections_parsed:
            self.directives_parsed = True
            self.selections_parsed = True
            return SelectionsParser(self.selections, None), 0

        if self.name is None or self.on_type is None:
            raise RuntimeError("Unexpected `None`")

        self.fragments.append(qm.NamedFragment(self.name, self.on_type, self.directives, self.selections))
        return None, 1

    def to_dbg_repr(self) -> PrimitiveType:
        d = t.cast(t.Dict[str, PrimitiveType], super().to_dbg_repr())

        add_if_not_none(d, "name", self.name)

        if self.on_type is not None:
            d["on_type"] = self.on_type.to_primitive()

        add_if_not_empty(d, "dirs", self.directives)
        add_if_not_empty(d, "sels", self.selections)

        return d


__all__ = ["parse_document", "parse", "ElementParser", "DocumentParser", "OperationParser", "VariablesParser",
           "VariableDefinitionParser", "DirectivesParser", "DirectiveParser", "ArgumentsParser",
           "ArgumentParser", "SelectionsParser", "FieldParser", "ValueParser", "ConstValueParser",
           "ListValueParser", "ConstListValueParser", "ObjectValueParser", "ConstObjectValueParser",
           "StringValueParser", "FragmentSpreadParser", "InlineFragmentParser", "FragmentParser"]



import typing as t


class Reader:
    text_input: str
    index: int
    lineno: int

    def __init__(self, text_input: str) -> None:
        self.text_input = text_input
        self.index = 0
        self.lineno = 1

    def str_read_ch(self) -> t.Optional[str]:
        if len(self.text_input) == self.index:
            return None

        ch = self.text_input[self.index]
        self.index += 1
        return ch

    def read_ch(self) -> t.Optional[str]:
        ch = self.lookup_ch()
        if ch is not None:
            self.index += 1
        return ch

    def lookup_ch(self) -> t.Optional[str]:
        while True:
            if len(self.text_input) == self.index:
                return None

            ch = self.text_input[self.index]

            if ch == "\n":
                self.lineno += 1
                self.index += 1
                continue

            if ch in " \r\t,":
                self.index += 1
                continue

            if ch == "#":
                index = self.text_input.find("\n", self.index)
                if index < 0:
                    self.index = len(self.text_input)
                else:
                    self.index = index
                continue

            return ch

    def match_re(self, regexp: t.Pattern[str]) -> t.Match[str]:
        return regexp.match(self.text_input, self.index)

    def read_re(self, regexp: t.Pattern[str]) -> t.Optional[str]:

        # eat ignoring characters
        self.lookup_ch()

        m = regexp.match(self.text_input, self.index)
        if m:
            self.index += len(m.group(0))
            return m.group(0)

        return None

    def line_pos(self) -> int:
        start = self.text_input.rfind("\n", 0, self.index)

        if start < 0:
            return self.index

        return self.index - start - 1

    def current_line(self) -> str:
        start = self.text_input.rfind("\n", 0, self.index)
        if start < 0:
            start = 0
        else:
            start += 1
        end = self.text_input.find("\n", self.index)
        if end < 0:
            end = len(self.text_input)
        return self.text_input[start:end]

    def prev_line(self) -> t.Optional[str]:
        prev_line_end = self.text_input.rfind("\n", 0, self.index)

        if prev_line_end < 0:
            return None

        if prev_line_end == 0:
            return ""

        prev_line_start = self.text_input.rfind("\n", 0, prev_line_end)

        if prev_line_start < 0:
            prev_line_start = 0
        else:
            prev_line_start += 1

        return self.text_input[prev_line_start:prev_line_end]

    def next_line(self) -> t.Optional[str]:
        next_line_start = self.text_input.find("\n", self.index)

        if next_line_start < 0:
            return None

        if next_line_start == len(self.text_input) - 1:
            return ""

        next_line_start += 1

        next_line_end = self.text_input.find("\n", next_line_start)

        if next_line_end < 0:
            next_line_end = len(self.text_input)

        return self.text_input[next_line_start:next_line_end]


def format_position(lineno: int, line_pos: int, lines: t.Sequence[t.Optional[str]]) -> t.Sequence[str]:
    result = []

    if lines[0] is not None:
        result.append("{:4d} | {}".format(lineno - 1, lines[0]))

    result.append("{:4d} | {}".format(lineno, lines[1]))
    result.append("       {}\u2303".format(" " * line_pos))

    if lines[2] is not None:
        result.append("{:4d} | {}".format(lineno + 1, lines[2]))

    return result


__all__ = ["Reader", "format_position"]



import json
import logging
import typing as t

import gql_alchemy.query_model as qm
import gql_alchemy.schema as s
import gql_alchemy.types as gt
from .errors import GqlValidationError
from .utils import PrimitiveType

logger = logging.getLogger("gql_alchemy")


class Env:
    def __init__(self, vars_definitions: t.Mapping[str, t.Union[gt.InputType, gt.WrapperType]],
                 vars_values: t.Optional[t.Mapping[str, PrimitiveType]]) -> None:
        self.vars_definitions = vars_definitions
        self.vars_values = vars_values


class Validator(qm.QueryVisitor):
    def __init__(self, type_registry: gt.TypeRegistry, query: gt.Object,
                 mutation: t.Optional[gt.Object]) -> None:
        self.__name = type(self).__name__
        self.type_registry = type_registry
        self.__query = query
        self.__mutation = mutation

        self._def_name = ""
        self._spreadables: t.List[gt.SpreadableType] = []
        self._location: gt.DirectiveLocation = gt.DirectiveLocations.QUERY

    def visit_query_begin(self, query: qm.Query) -> None:
        logger.debug("%s - visit query %s", self.__name, query.name)
        self._def_name = query.name if query.name is not None else "!"
        self._spreadables.append(self.__query)
        self._location = gt.DirectiveLocations.QUERY

    def visit_query_end(self, query: qm.Query) -> None:
        logger.debug("%s - finish query %s", self.__name, query.name)
        del self._spreadables[-1]

    def visit_mutation_begin(self, mutation: qm.Mutation) -> None:
        logger.debug("%s - visit mutation %s", self.__name, mutation.name)
        if self.__mutation is None:
            raise GqlValidationError("Mutations are not allowed by schema")

        self._def_name = mutation.name if mutation.name is not None else "!"
        self._spreadables.append(self.__mutation)
        self._location = gt.DirectiveLocations.MUTATION

    def visit_mutation_end(self, mutation: qm.Mutation) -> None:
        logger.debug("%s - finish mutation %s", self.__name, mutation.name)
        del self._spreadables[-1]

    def visit_fragment_begin(self, fragment: qm.NamedFragment) -> None:
        logger.debug("%s - visit fragment %s", self.__name, fragment.name)
        self._def_name = fragment.name

        on_type = self._resolve_type(fragment.on_type)
        spreadable = gt.is_spreadable(on_type)
        if spreadable is None:
            raise GqlValidationError(
                "Fragments can be defined only on spreadable types; problem with `{}` fragment".format(fragment.name)
            )
        self._spreadables.append(spreadable)

        self._location = gt.DirectiveLocations.FRAGMENT_DEFINITION

    def visit_fragment_end(self, fragment: qm.NamedFragment) -> None:
        logger.debug("%s - finish fragment %s", self.__name, fragment.name)
        del self._spreadables[-1]

    def visit_field_selection_begin(self, field_sel: qm.FieldSelection) -> None:
        logger.debug("%s - visit field %s", self.__name, field_sel.name)
        self._location = gt.DirectiveLocations.FIELD

        field_def = self.__get_field_def(field_sel)
        field_type = field_def.type(self.type_registry)

        if len(field_sel.selections) > 0:
            spreadable = gt.is_spreadable(self.type_registry.resolve_and_unwrap(field_type))
            if spreadable is None:
                raise GqlValidationError("Can not select from non spreadable type `{}`".format(field_type))
            self._spreadables.append(spreadable)
        else:
            spreadable = gt.is_spreadable(self.type_registry.resolve_and_unwrap(field_type))
            if spreadable is not None:
                raise GqlValidationError("Spreadable type `{}` must be selected with sub-selections".format(field_type))

    def visit_field_selection_end(self, field_sel: qm.FieldSelection) -> None:
        logger.debug("%s - finish field %s", self.__name, field_sel.name)
        if len(field_sel.selections) > 0:
            del self._spreadables[-1]

    def visit_fragment_spread_begin(self, spread: qm.FragmentSpread) -> None:
        logger.debug("%s - fragment spread %s", self.__name, spread.fragment_name)
        self._location = gt.DirectiveLocations.FRAGMENT_SPREAD

    def visit_inline_fragment_begin(self, fragment: qm.InlineFragment) -> None:
        logger.debug("%s - visit inline fragment", self.__name)
        self._location = gt.DirectiveLocations.INLINE_FRAGMENT

        if fragment.on_type is not None:
            on_type_name = fragment.on_type.name
            spreadable = self._spreadables[-1]
            if isinstance(spreadable, gt.Object):
                if on_type_name != str(spreadable):
                    raise GqlValidationError(
                        "Inline fragment on object field must be of the same type; `{}` expected, got `{}`".format(
                            spreadable, on_type_name
                        )
                    )
                self._spreadables.append(spreadable)
            elif isinstance(spreadable, gt.Union) or isinstance(spreadable, gt.Interface):
                of_objects = {str(o) for o in spreadable.of_objects(self.type_registry)}
                if on_type_name not in of_objects:
                    of_objects_list = list(of_objects)
                    of_objects_list.sort()
                    raise GqlValidationError("One of {} types expected, but got `{}`".format(
                        ", ".join(('`' + o + '`' for o in of_objects_list)),
                        on_type_name
                    ))
                self._spreadables.append(gt.assert_spreadable(self.type_registry.resolve_type(on_type_name)))
            else:
                raise RuntimeError("Spreadable must be Object, Interface or Union")
        else:
            if gt.is_selectable(self._spreadables[-1]) is None:
                raise GqlValidationError(
                    "Inline fragment without type possible only on selectable type, {} is not selectable".format(
                        self._spreadables[-1]
                    )
                )

    def visit_inline_fragment_end(self, fragment: qm.InlineFragment) -> None:
        logger.debug("%s - finish inline fragment", self.__name)
        if fragment.on_type is not None:
            del self._spreadables[-1]

    def _resolve_type(self, query_type: qm.Type) -> gt.GqlType:
        if isinstance(query_type, qm.NamedType):
            try:
                resolved_type = self.type_registry.resolve_type(query_type.name)
            except gt.TypeResolvingError as e:
                raise GqlValidationError("Unknown type `{}` used".format(query_type.name)) from e
            if query_type.null:
                return resolved_type
            return gt.NonNull(query_type.name)

        if isinstance(query_type, qm.ListType):
            element_type = self._resolve_type(query_type.el_type)

            inline = gt.is_inline(element_type)
            if inline is not None:
                list_type = gt.List(inline)
            else:
                list_type = gt.List(str(element_type))

            if query_type.null:
                return list_type

            return gt.NonNull(list_type)

        raise RuntimeError("One of named or list type expected here")

    def __get_field_def(self, field_sel: qm.FieldSelection) -> gt.Field:
        selectable = gt.is_selectable(self._spreadables[-1])
        if selectable is None:
            raise GqlValidationError("Can not select `{}` from non selectable `{}` type".format(
                field_sel.name, str(self._spreadables[-1])
            ))

        fields_defs = selectable.fields(self.type_registry)
        if field_sel.name not in fields_defs:
            raise GqlValidationError("`{}` type does not define `{}` field".format(str(selectable), field_sel.name))
        return fields_defs[field_sel.name]

    def _path(self) -> str:
        path_parts = [self._def_name]

        for sp in self._spreadables:
            path_parts.append(str(sp))

        return "/".join(path_parts)


class PassOne(Validator):
    def __init__(self, type_registry: gt.TypeRegistry, query: gt.Object, mutation: t.Optional[gt.Object]) -> None:
        super().__init__(type_registry, query, mutation)
        self.fragments: t.Dict[str, gt.SpreadableType] = {}

    def visit_fragment_begin(self, fragment: qm.NamedFragment) -> None:
        super().visit_fragment_begin(fragment)
        on_type = self._resolve_type(fragment.on_type)
        spreadable = gt.is_spreadable(on_type)
        if spreadable is None:
            raise GqlValidationError(
                "Fragment must be defined on spreadable type, `{}` is not spreadable".format(on_type)
            )
        self.fragments[fragment.name] = spreadable


class PassTwo(Validator):
    def __init__(self, fragments: t.Dict[str, gt.SpreadableType],
                 type_registry: gt.TypeRegistry, query: gt.Object, mutation: t.Optional[gt.Object],
                 vars_values: t.Mapping[str, PrimitiveType],
                 op_to_run: t.Optional[str]) -> None:
        super().__init__(type_registry, query, mutation)
        self.__fragments = fragments
        self.__vars_values = vars_values
        self.__op_to_run = op_to_run

        self.__root: t.Optional[t.Union[qm.Operation, qm.NamedFragment]] = None
        self.fragment_calls: t.Dict[str, t.List[t.Union[qm.Operation, qm.NamedFragment]]] = {}
        self.environments: t.Dict[str, Env] = {}

    def visit_query_begin(self, query: qm.Query) -> None:
        super().visit_query_begin(query)
        self.__root = query
        self.__save_env(query)

    def visit_mutation_begin(self, mutation: qm.Mutation) -> None:
        super().visit_mutation_begin(mutation)
        self.__root = mutation
        self.__save_env(mutation)

    def visit_fragment_begin(self, fragment: qm.NamedFragment) -> None:
        super().visit_fragment_begin(fragment)
        self.__root = fragment

    def visit_fragment_spread_begin(self, spread: qm.FragmentSpread) -> None:
        super().visit_fragment_spread_begin(spread)

        if self.__root is None:
            raise RuntimeError("Root object always expected here")

        if spread.fragment_name not in self.__fragments:
            raise GqlValidationError("Undefined fragment `{}` called".format(spread.fragment_name))

        called_on_type = self._spreadables[-1]
        defined_on_type = self.__fragments[spread.fragment_name]

        if str(called_on_type) != str(defined_on_type):
            if isinstance(called_on_type, gt.Interface) or isinstance(called_on_type, gt.Union):
                if str(defined_on_type) not in {str(o) for o in called_on_type.of_objects(self.type_registry)}:
                    raise GqlValidationError(
                        "Fragment `{}` can not be called on `{}` type".format(spread.fragment_name, called_on_type)
                    )
            else:
                raise GqlValidationError(
                    "Fragment `{}` can not be called on `{}` type".format(spread.fragment_name, called_on_type)
                )

        if spread.fragment_name not in self.fragment_calls:
            self.fragment_calls[spread.fragment_name] = []
        self.fragment_calls[spread.fragment_name].append(self.__root)

    def __save_env(self, op: qm.Operation) -> None:
        vars_definitions: t.Dict[str, t.Union[gt.InputType, gt.WrapperType]] = {}
        vars_defaults: t.Dict[str, t.Optional[qm.ConstValue]] = {}

        for var in op.variables:
            var_type = self._resolve_type(var.type)
            var_type_input_or_wrapper: t.Union[gt.InputType, gt.WrapperType, None] = gt.is_input(var_type)
            if var_type_input_or_wrapper is None:
                var_type_input_or_wrapper = gt.is_wrapper(var_type)
                if var_type_input_or_wrapper is None:
                    raise GqlValidationError(
                        "Only input types and their wrappers can be used as var types; "
                        "var `{}` of `{}` operation is not input type".format(var.name, op.name)
                    )
                var_type_input = gt.is_input(self.type_registry.resolve_and_unwrap(var_type_input_or_wrapper))
                if var_type_input is None:
                    raise GqlValidationError(
                        "Only input types and their wrappers can be used as var types; "
                        "var `{}` of `{}` operation is not input type".format(var.name, op.name)
                    )
                vars_definitions[var.name] = var_type_input_or_wrapper
            else:
                vars_definitions[var.name] = var_type_input_or_wrapper

            if var.default is not None:
                if not var_type_input_or_wrapper.validate_input(var.default, None, {}, self.type_registry):
                    raise GqlValidationError("Non compatible default value for `{}` variable of `{}` operation".format(
                        var.name, op.name
                    ))

            vars_defaults[var.name] = var.default

        if self.__is_running(op):
            vars_values = dict(self.__vars_values)
            for var_name, default in vars_defaults.items():
                if default is not None:
                    vars_values.setdefault(var_name, default.to_py_value({}))
            env = Env(vars_definitions, vars_values)
        else:
            env = Env(vars_definitions, None)

        for var_name, var_type in env.vars_definitions.items():
            default = vars_defaults[var_name]
            if default is not None:
                if not var_type.validate_input(default, None, {}, self.type_registry):
                    raise GqlValidationError(
                        "Variable can not be assigned to its default; "
                        "problem with `{}` variable in `{}` operation".format(var_name, op.name)
                    )
            if env.vars_values is not None:
                if var_name in env.vars_values:
                    if not var_type.is_assignable(env.vars_values[var_name], self.type_registry):
                        raise GqlValidationError(
                            "Wrong value {} provided for `{}` variable of `{}` operation".format(
                                json.dumps(env.vars_values[var_name]), var_name, op.name
                            )
                        )
                elif vars_defaults[var_name] is None:
                    raise GqlValidationError(
                        "Variable `{}` is required in `{}` operation".format(var_name, op.name)
                    )

        self.environments[op.name if op.name is not None else "!"] = env

    def __is_running(self, op: qm.Operation) -> bool:
        if self.__op_to_run is None:
            return True

        return op.name == self.__op_to_run


class PassThree(Validator):
    """Verify operations and collect environments for fragments used in query"""

    def __init__(self,
                 environments: t.Mapping[str, Env],
                 fragment_calls: t.Mapping[str, t.Sequence[t.Union[qm.Operation, qm.NamedFragment]]],
                 type_registry: gt.TypeRegistry,
                 query: gt.Object,
                 mutation: t.Optional[gt.Object]) -> None:
        Validator.__init__(self, type_registry, query, mutation)
        self.__environments = environments
        self.__fragment_calls = fragment_calls

        self.__current_envs: t.Sequence[Env] = []
        self.__args_defs: t.Mapping[str, gt.Argument] = {}

    def visit_query_begin(self, query: qm.Query) -> None:
        super().visit_query_begin(query)
        self.__current_envs = [self.__environments[self._def_name]]

    def visit_mutation_begin(self, mutation: qm.Mutation) -> None:
        super().visit_mutation_begin(mutation)
        self.__current_envs = [self.__environments[self._def_name]]

    def visit_fragment_begin(self, fragment: qm.NamedFragment) -> None:
        super().visit_fragment_begin(fragment)
        self.__current_envs = self.__resolve_fragment_envs(fragment)

    def visit_field_selection_begin(self, field_sel: qm.FieldSelection) -> None:
        super().visit_field_selection_begin(field_sel)

        if len(field_sel.selections) > 0:
            parent_spreadable = self._spreadables[-2]
        else:
            parent_spreadable = self._spreadables[-1]

        parent_selectable = gt.assert_selectable(parent_spreadable)
        field_def = parent_selectable.fields(self.type_registry)[field_sel.name]

        self.__validate_args(field_sel.arguments, field_def.args)
        self.__args_defs = field_def.args

    def visit_directive_begin(self, directive: qm.Directive) -> None:
        super().visit_directive_begin(directive)

        dir_def = self.__resolve_directive(directive.name)
        if self._location not in dir_def.locations:
            raise GqlValidationError("Can not use `{}` in `{}` location".format(directive.name, self._location))

        self.__validate_args(directive.arguments, dir_def.args)
        self.__args_defs = dir_def.args

    def visit_argument(self, argument: qm.Argument) -> None:
        super().visit_argument(argument)

        if argument.name not in self.__args_defs:
            raise GqlValidationError("Unsupported argument `{}`".format(argument.name))

        arg_def = self.__args_defs[argument.name]

        for env in self.__current_envs:
            try:
                if not arg_def.validate_input(argument.value, env.vars_values, env.vars_definitions,
                                              self.type_registry):
                    raise GqlValidationError("Can not use `{}` as `{}` argument".format(
                        json.dumps(argument.value.to_primitive()),
                        argument.name
                    ))
            except gt.UndefinedVariableError as e:
                raise GqlValidationError("{}; ??".format(
                    str(e)
                )) from e
            except gt.NonCompatibleVariableType as e:
                raise GqlValidationError("{}; ??".format(
                    str(e)
                )) from e

    def __validate_args(self, args_provided: t.Sequence[qm.Argument], args_defs: t.Mapping[str, gt.Argument]) -> None:
        args_defined_names = {a for a in args_defs.keys()}
        args_required_names = {
            name for name, arg in args_defs.items()
            if arg.default is None and isinstance(arg.type(self.type_registry), gt.NonNull)
        }
        args_provided_names = {arg.name for arg in args_provided}

        unsupported_args = list(args_provided_names.difference(args_defined_names))
        if len(unsupported_args) > 0:
            if len(unsupported_args) == 1:
                raise GqlValidationError("Argument `{}` is not supported".format(
                    next(iter(unsupported_args))
                ))
            else:
                unsupported_args.sort()
                raise GqlValidationError("Arguments {} are not supported".format(
                    ', '.join(("`" + arg_name + "`" for arg_name in unsupported_args))
                ))

        unfilled_args = list(args_required_names.difference(args_provided_names))
        if len(unfilled_args) > 0:
            if len(unfilled_args) == 1:
                raise GqlValidationError("Argument `{}` is required".format(
                    next(iter(unfilled_args))
                ))
            else:
                unfilled_args.sort()
                raise GqlValidationError("Arguments {} are required".format(
                    ', '.join(("`" + arg_name + "`" for arg_name in unfilled_args))
                ))

    def __resolve_directive(self, name: str) -> gt.Directive:
        try:
            return self.type_registry.directive(name)
        except gt.TypeResolvingError as e:
            raise GqlValidationError("Unknown directive `{}` used".format(name)) from e

    def __resolve_fragment_envs(self, fragment: qm.NamedFragment) -> t.Sequence[Env]:
        if fragment.name not in self.__fragment_calls:
            raise GqlValidationError("Unused fragment `{}`".format(fragment.name))

        envs = []

        for called_from in self.__fragment_calls[fragment.name]:
            if isinstance(called_from, qm.NamedFragment):
                for env in self.__resolve_fragment_envs(called_from):
                    envs.append(env)
            else:
                envs.append(self.__environments[called_from.name if called_from.name is not None else "!"])

        return envs


def validate(query: qm.Document, schema: s.Schema,
             vars_values: t.Mapping[str, PrimitiveType], op_to_run: t.Optional[str] = None) -> None:
    if op_to_run is None and len(query.operations) > 1:
        raise GqlValidationError("You must specify query to run for queries with many operations")

    if op_to_run is not None and op_to_run not in {op.name for op in query.operations}:
        raise GqlValidationError("Operation requested to run is not defined in the query")

    type_registry = schema.type_registry

    query_obj = type_registry.resolve_type(schema.query_object_name)
    if not isinstance(query_obj, gt.Object):
        raise RuntimeError("Query must be an object")

    mutation_obj = None
    if schema.mutation_object_name is not None:
        mutation_obj = type_registry.resolve_type(schema.mutation_object_name)
    if mutation_obj is not None and not isinstance(mutation_obj, gt.Object):
        raise RuntimeError("Mutation must be an object")

    pass_one = PassOne(type_registry, query_obj, mutation_obj)
    query.visit(pass_one)

    pass_two = PassTwo(pass_one.fragments, type_registry, query_obj, mutation_obj, vars_values, op_to_run)
    query.visit(pass_two)

    pass_three = PassThree(pass_two.environments, pass_two.fragment_calls, type_registry, query_obj, mutation_obj)
    query.visit(pass_three)



import json
import typing as t

import gql_alchemy.query_model as qm
import gql_alchemy.schema as s
import gql_alchemy.types as gt
from .errors import GqlExecutionError
from .parser import parse_document
from .resolvers import Resolver, IntrospectionResolver, Introspection
from .utils import PrimitiveType
from .validator import validate

_py_reserved = {
    "False", "class", "finally", "is", "return",
    "None", "continue", "for", "lambda", "try",
    "True", "def", "from", "nonlocal", "while",
    "and", "del", "global", "not", "with",
    "as", "elif", "if", "or", "yield",
    "assert", "else", "import", "pass",
    "break", "except", "in", "raise"
}


SomeResolver = t.TypeVar('SomeResolver', bound=Resolver)


class Directive:
    def should_select_field(self, resolver: SomeResolver,
                            field_name: str) -> bool:
        return True

    def wrap_field(self, field: t.Any,
                   field_args: t.Mapping[str, PrimitiveType]) -> t.Any:
        return field


class _SkipDirective(Directive):
    def __init__(self, _if: bool) -> None:
        self.__if = _if

    def should_select_field(self, resolver: SomeResolver,
                            field_name: str) -> bool:
        return not self.__if


class _IncludeDirective(Directive):
    def __init__(self, _if: bool) -> None:
        self.__if = _if

    def should_select_field(self, resolver: SomeResolver,
                            field_name: str) -> bool:
        return self.__if


class _DirectivesEnv:
    def __init__(self, parent_directives: t.MutableSet[Directive],
                 own_directives: t.Sequence[qm.Directive],
                 directives_constructors: t.Mapping[str, t.Callable[..., Directive]],
                 vars_values: t.Mapping[str, PrimitiveType],
                 type_registry: gt.TypeRegistry) -> None:
        self.__parent_directives = parent_directives
        self.__own_directives = own_directives
        self.__type_registry = type_registry
        self.__directives_constructors = directives_constructors
        self.__vars_values = vars_values
        self.__added_directives: t.MutableSet[Directive] = set()

    def __enter__(self) -> None:
        for d in self.__own_directives:
            dir_c = self.__directives_constructors[d.name]
            d_def = self.__type_registry.directive(d.name)
            args = _prepare_args(d.arguments, self.__vars_values, d_def.args)
            directive = dir_c(**args)
            self.__parent_directives.add(directive)
            self.__added_directives.add(directive)

    def __exit__(self, exc_type: t.Any, exc_val: t.Any, exc_tb: t.Any) -> None:
        for d in self.__added_directives:
            self.__parent_directives.remove(d)


class Executor:
    def __init__(self, schema: s.Schema, query_resolver: SomeResolver,
                 mutation_resolver: t.Optional[SomeResolver] = None,
                 directives: t.Optional[t.Mapping[str, t.Callable[..., Directive]]] = None) -> None:
        self.schema = schema
        self.type_registry = schema.type_registry
        self.query_resolver = query_resolver
        self.mutation_resolver = mutation_resolver
        self.query_object_name = schema.query_object_name
        self.mutation_object_name = schema.mutation_object_name
        self.directives: t.Dict[str, t.Callable[..., Directive]] = dict(directives) if directives is not None else {}
        self.directives["skip"] = _SkipDirective
        self.directives["include"] = _IncludeDirective

        if self.mutation_object_name is not None and mutation_resolver is None:
            raise GqlExecutionError("Mutation resolver required with schema that supports mutations")

    def query(self, query: str, variables: t.Mapping[str, PrimitiveType],
              op_to_run: t.Optional[str] = None) -> PrimitiveType:
        document = parse_document(query)

        validate(document, self.schema, variables, op_to_run)

        if op_to_run is None and len(document.operations) > 1:
            raise GqlExecutionError("Operation name is needed for queries with multiple operations defined")

        operation: t.Optional[qm.Operation] = None
        if op_to_run is None:
            operation = document.operations[0]
        else:
            for op in document.operations:
                if op.name == op_to_run:
                    operation = op

        if operation is None:
            raise GqlExecutionError("Operation `{}` is not found".format(op_to_run))

        if isinstance(operation, qm.Query):
            root_object = self.type_registry.resolve_type(self.query_object_name)
            resolver: Resolver = IntrospectionResolver(self.query_resolver, Introspection(self.schema))
        else:
            if self.mutation_object_name is None or self.mutation_resolver is None:
                raise GqlExecutionError("Server does not support mutations")
            root_object = self.type_registry.resolve_type(self.mutation_object_name)
            resolver = self.mutation_resolver

        return _OperationRunner(self.type_registry, variables, document, self.directives).run_operation(
            t.cast(gt.Object, root_object),
            operation,
            resolver
        )


class _OperationRunner:
    def __init__(self, type_registry: gt.TypeRegistry,
                 vars_values: t.Mapping[str, PrimitiveType],
                 document: qm.Document,
                 directives: t.Mapping[str, t.Callable[..., Directive]]) -> None:
        self.type_registry = type_registry
        self.vars_values = dict(vars_values)
        self.fragments = dict(((f.name, f) for f in document.fragments))
        self.directives = directives

    def run_operation(self, root_object: gt.Object, operation: qm.Operation,
                      root_resolver: Resolver) -> t.Mapping[str, PrimitiveType]:
        for var in operation.variables:
            if var.default is not None:
                self.vars_values.setdefault(var.name, var.default.to_py_value({}))

        result: t.Dict[str, PrimitiveType] = {}
        directives: t.MutableSet[Directive] = set()

        with _DirectivesEnv(directives, operation.directives, self.directives, self.vars_values, self.type_registry):
            self.__select(directives, result, operation.selections, root_object, root_resolver)
        return result

    def __select(self, parent_directives: t.MutableSet[Directive], result: t.Dict[str, PrimitiveType],
                 selections: t.Sequence[qm.Selection],
                 from_selectable: gt.SpreadableType,
                 resolver: Resolver) -> None:
        for sel in selections:
            if isinstance(sel, qm.FieldSelection):
                selectable = gt.assert_selectable(from_selectable)
                field = selectable.fields(self.type_registry)[sel.name]
                self.__select_field(parent_directives, result, sel, field, resolver)
                continue
            if isinstance(sel, qm.FragmentSpread):
                with _DirectivesEnv(parent_directives, sel.directives, self.directives, self.vars_values,
                                    self.type_registry):
                    self.__select_fragment(parent_directives, result, self.fragments[sel.fragment_name],
                                           from_selectable,
                                           resolver)
            if isinstance(sel, qm.InlineFragment):
                self.__select_fragment(parent_directives, result, sel, from_selectable, resolver)

    def __select_fragment(self, parent_directives: t.MutableSet[Directive],
                          result: t.Dict[str, PrimitiveType], frg: qm.Fragment,
                          from_selectable: gt.SpreadableType,
                          resolver: SomeResolver) -> None:
        with _DirectivesEnv(parent_directives, frg.directives, self.directives, self.vars_values, self.type_registry):
            if frg.on_type is not None:
                on_type = gt.assert_spreadable(self.__resolve_type(frg.on_type.name))
                if isinstance(on_type, gt.Interface) or isinstance(on_type, gt.Union):
                    possible_objects = {str(o) for o in on_type.of_objects(self.type_registry)}
                else:
                    if not isinstance(on_type, gt.Object):
                        raise RuntimeError("Object expected here")
                    possible_objects = {str(on_type)}

                if resolver.for_gql_type not in possible_objects:
                    return

                resolver_type = gt.assert_selectable(self.__resolve_type(resolver.for_gql_type))

                self.__select(parent_directives, result, frg.selections, resolver_type, resolver)

            else:
                self.__select(parent_directives, result, frg.selections, from_selectable, resolver)

    def __select_field(self, parent_directives: t.MutableSet[Directive], result: t.Dict[str, PrimitiveType],
                       field_selection: qm.FieldSelection,
                       field_definition: gt.Field,
                       resolver: SomeResolver) -> None:
        with _DirectivesEnv(parent_directives, field_selection.directives, self.directives, self.vars_values,
                            self.type_registry):
            for d in parent_directives:
                if not d.should_select_field(resolver, field_selection.name):
                    return

            args = _prepare_args(field_selection.arguments, self.vars_values, field_definition.args)

            field_type = field_definition.type(self.type_registry)
            alias = field_selection.alias if field_selection.alias is not None else field_selection.name

            field_name = field_selection.name
            if field_name in _py_reserved:
                field_name = "_" + field_name
            elif field_name.startswith("__"):
                field_name = "f" + field_name

            if not hasattr(resolver, field_name):
                raise GqlExecutionError("Resolver `{}` for `{}` type does not have `{}` attribute".format(
                    type(resolver).__name__, resolver.for_gql_type, field_name
                ))

            attr = getattr(resolver, field_name)
            for d in parent_directives:
                attr = d.wrap_field(attr, args)

            if len(args) == 0 and not callable(attr):
                field_raw_value = attr
            else:
                try:
                    field_raw_value = attr(**args)
                except Exception as e:
                    raise GqlExecutionError("Resolver internal error") from e

            if len(field_selection.selections) > 0:
                result[alias] = self.__process_spreadable_field(
                    parent_directives, field_type, field_raw_value, field_selection.selections
                )
            else:
                result[alias] = self.__select_plain_field(resolver, field_name, field_type, field_raw_value)

    def __process_spreadable_field(self, parent_directives: t.MutableSet[Directive], field_type: gt.GqlType,
                                   field_raw_value: t.Any,
                                   selections: t.Sequence[qm.Selection]) -> PrimitiveType:

        if not self.__resolver_compatible(field_type, field_raw_value):
            raise GqlExecutionError("Resolver returns non compatible sub-resolver")

        return self.__resolve_subresolvers(
            parent_directives,
            selections,
            gt.assert_spreadable(self.__resolve_and_unwrap(field_type)),
            field_raw_value
        )

    def __resolver_compatible(self, field_type: gt.GqlType, field_raw_value: t.Any) -> bool:
        if isinstance(field_type, gt.NonNull):
            if field_raw_value is None:
                return False
            return self.__resolver_compatible(field_type.of_type(self.type_registry), field_raw_value)

        if field_raw_value is None:
            return True

        if isinstance(field_type, gt.List):
            if not isinstance(field_raw_value, list):
                return False
            for resolver in field_raw_value:
                if not self.__resolver_compatible(field_type.of_type(self.type_registry), resolver):
                    return False
            return True

        if isinstance(field_type, gt.Interface) or isinstance(field_type, gt.Union):
            possible_objects = field_type.of_objects(self.type_registry)
            possible_objects_names = {str(o) for o in possible_objects}
            return field_raw_value.for_gql_type in possible_objects_names

        if isinstance(field_type, gt.Object):
            resolver_type = t.cast(t.Optional[str], field_raw_value.for_gql_type)
            return str(field_type) == resolver_type

        raise RuntimeError("Wrapper or spreadable expected here, but got {}".format(type(field_type).__name__))

    def __resolve_subresolvers(self, parent_directives: t.MutableSet[Directive], selections: t.Sequence[qm.Selection],
                               field_type: gt.SpreadableType,
                               field_raw_value: t.Any) -> PrimitiveType:
        if field_raw_value is None:
            return None

        if isinstance(field_raw_value, list):
            result_arr: t.List[PrimitiveType] = []
            for item_raw_value in field_raw_value:
                result_arr.append(
                    self.__resolve_subresolvers(parent_directives, selections, field_type, item_raw_value)
                )
            return result_arr

        result_dict: t.Dict[str, PrimitiveType] = {}
        self.__select(parent_directives, result_dict, selections, field_type, field_raw_value)
        return result_dict

    def __select_plain_field(self, resolver: Resolver, field_name: str, field_type: gt.GqlType,
                             field_raw_value: t.Any) -> PrimitiveType:
        if not field_type.is_assignable(field_raw_value, self.type_registry):
            raise GqlExecutionError(
                "Resolver `{}` for type `{}` returns not assignable value '{}' for field `{}` of type `{}`".format(
                    type(resolver).__name__, resolver.for_gql_type, json.dumps(field_raw_value), field_name,
                    str(field_type)
                )
            )

        return t.cast(PrimitiveType, field_raw_value)

    def __resolve_type(self, schema_type: t.Union[gt.GqlType, str]) -> gt.GqlType:
        return self.type_registry.resolve_type(schema_type)

    def __resolve_and_unwrap(self, schema_type: t.Union[gt.GqlType, str]) -> gt.NonWrapperType:
        return self.type_registry.resolve_and_unwrap(schema_type)


def _prepare_args(arguments: t.Sequence[qm.Argument],
                  vars_values: t.Mapping[str, PrimitiveType],
                  args_def: t.Mapping[str, gt.Argument]) -> t.Mapping[str, PrimitiveType]:
    args_values = {}
    for arg in arguments:
        name = arg.name if arg.name not in _py_reserved else "_" + arg.name
        args_values[name] = arg.value.to_py_value(vars_values)

    for arg_name, arg_def in args_def.items():
        name = arg_name if arg_name not in _py_reserved else "_" + arg_name
        args_values.setdefault(name, arg_def.default)

    return args_values



import typing as t

from .raw_reader import Reader, format_position


class GqlError(Exception):
    pass


class GqlParsingError(GqlError):
    """Errors during query parsing"""

    def __init__(self, msg: str, reader: t.Optional[Reader] = None) -> None:
        self.msg = msg

        self.lineno: t.Optional[int] = None
        self.line_pos: t.Optional[int] = None
        self.lines: t.Optional[t.Sequence[t.Optional[str]]] = None

        if reader is not None:
            self.lineno = reader.lineno
            self.line_pos = reader.line_pos()
            self.lines = [reader.prev_line(), reader.current_line(), reader.next_line()]

    def __str__(self) -> str:
        lines = [self.msg]

        if self.lineno is not None and self.line_pos is not None and self.lines is not None:
            lines += format_position(self.lineno, self.line_pos, self.lines)

        return '\n'.join(lines)


class GqlValidationError(GqlError):
    pass


class GqlSchemaError(GqlError):
    """Errors in schema definition"""
    pass


class GqlExecutionError(GqlError):
    """Errors during execution: response do not match schema"""
    pass


__all__ = ["GqlError", "GqlParsingError", "GqlSchemaError", "GqlExecutionError",
           "GqlValidationError"]



from setuptools import setup, find_packages

setup(
    name="gql-alchemy",
    version="0.0.1",
    description="GraphQL implementation",
    long_description="Implementation for Facebook GraphQL based on http://facebook.github.io/graphql/October2016/ spec",
    url="https://github.com/gql-alchemy/gql-alchemy",
    author="Dmitry Maslennikov",
    author_email="maslennikovdm+gql-alchemy@gmail.com",
    license="MIT",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3.6',
    ],
    keywords="graphql",
    packages=find_packages(exclude=["tests"]),
    python_requires="~=3.6"
)
